;;; ns-grabenv.el --- functions to set environment variables by running a subshell
;;; Code taken from 4.2 NS distribution.
;;; Experimental.

;;;###autoload
(defun ns-grabenv (&optional shell-path &rest startup)
  "Run a shell subprocess, and interpret its output as a series of environment\n\
variables to insert into the emacs environment.  The first optional argument\n\
gives the path to the shell (defaults to \"/bin/csh\").  The remaining arguments\n\
are interpreted as a list of commands for it to execute (defaults to\n\
\"source /etc/csh.login\", \"source ~/.login\" and \"setenv\")."
  (interactive)
  (let ((tmpbuf (get-buffer-create " tmp-env"))
        (startlist (if startup startup '("source /etc/csh.login"
                                         "source ~/.login"
                                         "setenv"))))
    (save-excursion
     (set-buffer tmpbuf)
     (erase-buffer)
     (while startlist
       (insert (car startlist) "\n")
       (setq startlist (cdr startlist)))
     (call-process-region (point-min) (point-max)
                          (if shell-path shell-path "/bin/csh")
                          t t nil)
     (goto-char (point-min))
     (while (< (point) (point-max))
       (if (looking-at "^\\([^=\n]*\\)=\\(.*\\)$")
           (setenv (buffer-substring (match-beginning 1) (match-end 1))
                   (buffer-substring (match-beginning 2) (match-end 2))))
       (forward-line 1))
     (kill-buffer tmpbuf))))

(provide 'ns-grabenv)

;;; ns-grabenv.el ends here

; (autoload (quote ns-grabenv) "ns-grabenv" "\
; Run a shell subprocess, and interpret its output as a series of environment
; variables to insert into the emacs environment.  The first optional argument
; gives the path to the shell (defaults to \"/bin/csh\").  The remaining arguments
; are interpreted as a list of commands for it to execute (defaults to
; \"source /etc/login.std\", \"source ~/.login\" and \"setenv\")." nil nil)
