// from nsterm.h
#define NS_FONT_WIDTH(f)	((f)->output_data.ns->font->width)

// old ns_dynamic_path in nsterm.m
char *
ns_dynamic_path (char *startingPath)
// ----------------------------------------------------------------------------
// Used to allow emacs to find its resources under Emacs.app
// Called from callproc.c, lread.c
// We may replace this with Carbon's approach, which also allows resources
// to be buried less deeply
// ----------------------------------------------------------------------------
{
  NSMutableString *initPath =
      [NSMutableString stringWithUTF8String: startingPath];
  NSString *prefixDir;
  NSString *invDir =
      [NSString stringWithUTF8String: XSTRING(Vinvocation_directory)->data];
  NSRange range, searchRange;
  int pos;
  char *ibuf, *buffer;

  // get base
  range = [initPath rangeOfString: @"/Resources"];
  pos = range.location;
  if (pos == NSNotFound)
      return startingPath;
  prefixDir = [initPath substringToIndex: pos + range.length];

  // go back until either contents or .app
  range = [invDir rangeOfString: @"Contents"];
  if (range.location == NSNotFound)
      range = [invDir rangeOfString: @".app"];
  pos = range.location;
  if (pos == NSNotFound)
      return startingPath;
  pos += range.length;
  invDir = [invDir substringToIndex: pos + 1];
  invDir = [invDir stringByAppendingString: @"Resources"];

//NSLog(@"Replacing '%@' with '%@'", prefixDir, invDir);
  for (searchRange = NSMakeRange(0, [initPath length]); searchRange.length>0; )
    {
      range= [initPath rangeOfString: prefixDir options: 0 range: searchRange];
      if (range.location == NSNotFound)
        break;
      [initPath replaceCharactersInRange: range withString: invDir];
      searchRange.location = range.location + [invDir length];
      searchRange.length = [initPath length] - searchRange.location;
    }
  // method not on OS X 10.1
//    [initPath replaceOccurrencesOfString: prefixDir withString: invDir
//        options: 0 range: NSMakeRange(0, [initPath length])];

  ibuf = [initPath UTF8String];
  buffer = (char *)xmalloc((strlen(ibuf)+1)*sizeof(char));
  bcopy(ibuf, buffer, strlen(ibuf)+1);
//  fprintf(stderr, "dynPath start : '%s'\n",startingPath);
//  fprintf(stderr, "dynPath result: '%s'\n",buffer);
  return buffer;
}


// fragment from close-to-release nsterm.m ns_dumpglyphs() before background logic
// was simplified
  // Dump background in cases not picked up by ns_maybe_dumpglyphs_background
  // PENDING: look into consolidating.
  if ((flags == NS_DUMPGLYPH_NORMAL || flags == NS_DUMPGLYPH_MOUSEFACE)
      && ((/*!s->for_overlaps &&*/ !s->background_filled_p /*&& s->hl!=DRAW_CURSOR*/)
      //PENDING: w/o this line# drawing gets messed up, however with it there
      // are occasional obliterations in modeline (not as annoying though).
      // Probably due to a bug elsewhere (maybe lbearing/rbearing handling.)
      || s->row->mode_line_p))


// 23: no longer track mouse face ourselves..
// make the tracking rects used for mouse-face highlighting visible
#if 0
#define TRACKREC(r, v) {NSRect s=r; [v lockFocus]; NSHighlightRect(s);  \
  s.size.width-=2; s.size.height-=2; s.origin.x+=1; s.origin.y+=1;      \
  NSHighlightRect(s); trackrec=r; [v unlockFocus];                      \
  [[v window] flushWindow]; fprintf(stderr, "trackrec: %.0f, %.0f : %.0f x %.0f",r.origin.x,r.origin.y,r.size.width,r.size.height);}
#else
#define TRACKREC(r, v)
#endif


//23: this was causing problems, and Emacs now requests cursors anyway
- (void)resetCursorRects
{
  NSRect visible = [self visibleRect];
  NSPoint p = [[self window] mouseLocationOutsideOfEventStream];
  p = [self convertPoint:p fromView:nil];

  // adjust for scrollbar, which overlaps this
  visible.origin.x +=
    (FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(emacsframe) ? 1 : -1)
    * NS_SCROLL_BAR_WIDTH_DEFAULT;
  NSTRACE(resetCursorRects);

  if (!NSIsEmptyRect(visible))
    {
      [self addCursorRect: visible cursor: FRAME_POINTER_TYPE(emacsframe)];
      if (NSPointInRect(p, visible))
        [FRAME_POINTER_TYPE(emacsframe) set];
NSLog(@"ResetCursorRects w/cursor %@, ptinRect = %d",FRAME_POINTER_TYPE(emacsframe),NSPointInRect(p, visible));
    }
  [FRAME_POINTER_TYPE(emacsframe) setOnMouseEntered: YES];
}


//23: This function now only responsible for 1 face, so the code gets
//    MUCH simpler..

void
ns_dumpglyphs (FRAME_PTR nf, int x, int y, struct glyph *glyphs, int len,
               int flags)
// ----------------------------------------------------------------------------
//   Write 'len' glyphs to target frame 'nf' at position 'x,y'.
//   flags specifies drawing mode: normal, background, highlight, mouseface,
//   or foreground (last is NS-specific).
// ----------------------------------------------------------------------------
{
  static int *t[256];   // each element points to run of chars
  static unsigned int l[256];     // length of run
  static int buf[4096]; // holds chars in multiple runs
  static NSRect r[256];           // bounding rects of char run
  static int faces[256];
  static NSRect cRect;            // bounding rect for whole run
  static NSRect rectList[24];

  int *p=buf;
  int no = 0;
  int faceid;
  struct glyph cg;
  BOOL break_on_face_change = NO;
  BOOL left = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(nf);
  int firstCharPos = FIRST_CHAR_POSITION(nf);

  NSTRACE(dumpglyph);

  if (flags == NS_DUMPGLYPH_FOREGROUND
        && [NS_FACE_BACKGROUND(FRAME_DEFAULT_FACE(nf)) alphaComponent] != 1.0)
      return; // overdrawing foreground doesn't work well w/transparency

  cRect = NSMakeRect(0,0,0,0);
//if (flags == NS_DUMPGLYPH_NORMAL) fprintf(stderr, "%d chars starting with %d (%c), %d (%c)\n",len,glyphs[0],FAST_GLYPH_CHAR(glyphs[0]),glyphs[1],FAST_GLYPH_CHAR(glyphs[1]));

  while (len > 0)
    {
      // Select face based on input flags
      switch(flags)
        {
        case NS_DUMPGLYPH_HIGHLIGHT:
          faceid = 1;
          break;
        case NS_DUMPGLYPH_CURSOR:
        case NS_DUMPGLYPH_BACKGROUND:
          faceid = 0;
          break;
        case NS_DUMPGLYPH_MOUSEFACE:
          faceid = FRAME_NS_DISPLAY_INFO(nf)->mouse_face_face_id;
          break;
        default:
          g = *glyphs;
          GLYPH_FOLLOW_ALIASES (GLYPH_TABLE_BASE, GLYPH_TABLE_LENGTH, g);
          faceid = cg->faceid;
          break_on_face_change = YES;
          break;
        }

      faces[no] = faceid;

        // Add glyphs to current run of characters
      if (flags != NS_DUMPGLYPH_BACKGROUND)
        {
          int spaceRun = 0;
          for (t[no] = p; len > 0; len--)
            {
              g = *glyphs;
              GLYPH_FOLLOW_ALIASES (GLYPH_TABLE_BASE, GLYPH_TABLE_LENGTH, g);
              if (break_on_face_change && FAST_GLYPH_FACE(g) != faceid)
                break;          // End of run if face changes
              ///HACK to keep left scrollbar from being overwritten
              if (!left || (x + p-t[no]) >= firstCharPos || faceid == 1)
                {
                  *p++ = FAST_GLYPH_CHAR(g);
                  // HACK to add overhang when rendering engine clears text
                  // by rendering spaces where words were (aargh)
                  if (isspace(FAST_GLYPH_CHAR(g)))
                    spaceRun++;
                  else
                    spaceRun = 0;
                }
              else
                {
                  x++;
                }
              glyphs++;
            }
          if (p - t[no] > 0)
            {
              ns_chars_to_rect(nf, x, y, x + (p - t[no]) - 1, y, &r[no]);
              if (spaceRun > 1)
                r[no].size.width += 3;  // HACK: see above
              cRect = NSUnionRect(cRect, r[no]);
              l[no] = p - t[no];
              x += l[no];
              *p = '\0';
              no++;
            }
        }
      else // NS_DUMPGLYPH_BACKGROUND
        {
          ///HACK to keep scrollbar from being overwritten until the real
          //  cause of the problem can be found
          // (added for GNUstep but doesn't hurt OS X)
          if (left && (x < firstCharPos) && faceid != 1)
            x = firstCharPos;
          ns_chars_to_rect(nf, x, y, x + len - 1, y, &r[no]);
          cRect = NSUnionRect(cRect, r[no]);
          t[no] = p;
          l[no] = 0;
          len = 0;
          no++;
        }
      if (no >= 256)
        fprintf(stderr, "ASSERTION FAILURE! no > 256!\n");
    } // while glyphs remaining


  // Display assembled glyph runs
  if (no > 0)
    {
      int i;

      ns_focus(nf, &cRect);

      // Draw the text itself
      for (i = 0; i < no; i++)
        {

          // sort out face
          int face_idx = ((faces[i] < -1 ||
                           faces[i] >= FRAME_N_COMPUTED_FACES(nf) ||
                           FRAME_COMPUTED_FACE(nf, faces[i]) == 0)
                          ? 0
                          : faces[i]);
          struct face *face = FRAME_COMPUTED_FACE(nf, face_idx);
          XFontStruct *font = face->font == ~0 ? FRAME_FONT(nf) : face->font;

          // Fill background color unless drawing over cursor or fg-only
          if (flags!=NS_DUMPGLYPH_CURSOR && flags!=NS_DUMPGLYPH_FOREGROUND)
            {
              EmacsView *view = FRAME_NS_VIEW(nf);
              NSRectArray sbRects = view->sbRects;
              int nsbRects = view->nsbRects;

              if (!face->stipple)
                [(id)(NS_FACE_BACKGROUND(face)!=nil?
                  NS_FACE_BACKGROUND(face):FRAME_BACKGROUND_COLOR(nf)) set];
              else
                [[NSColor colorWithPatternImage:
                            [(id)face->stipple stippleRep]] set];

              if (nsbRects > 1 && face_idx != 1)
                {
                  int j;
                  int n=0;
                  NSRect tRect = r[i];
                  NSRect iRect;
                  for (j=0; j<nsbRects; j++)
                    {
                      NSRect sRect = sbRects[j];
                      //HACK
#ifdef COCOA
                      sRect.origin.x -= 2; sRect.size.height += 34; sRect.origin.y -= 34;
#else
                      sRect.origin.x -= 1; sRect.size.height += 32; sRect.origin.y -= 32;
#endif
                      iRect = NSIntersectionRect(sRect, tRect);
                      if (!NSIsEmptyRect(iRect))
                        {
                          NSRect before = tRect, after = tRect;
                          before.size.width = iRect.origin.x - tRect.origin.x;
                          after.origin.x = iRect.origin.x + iRect.size.width;
                          after.size.width -=
                            iRect.origin.x+iRect.size.width - tRect.origin.x;
                          if (!NSIsEmptyRect(before))
                            {
                              rectList[n++] = before;
//if (!NSIsEmptyRect(NSIntersectionRect(before,sbRects[j]))) fprintf(stderr,"NONEMPTY (1)!\n");
                            }
                          tRect = after;
                        }
                    }
                  if (!NSIsEmptyRect(tRect))
                    {
                      rectList[n++] = tRect;
//if (!NSIsEmptyRect(NSIntersectionRect(tRect,sbRects[0]))) fprintf(stderr,"NONEMPTY (2)!\n");
                    }
//fprintf(stderr, "filling %d rects...\n",n);
//                      { int k,l;
//                      for (k=0; k<n; k++)
//                          for (l=0; l<nsbRects; l++)
//                              if (!NSIsEmptyRect(NSIntersectionRect(rectList[k],sbRects[l]))) fprintf(stderr,"NONEMPTY: %d, %d!\n",k,l);
//                      }
                  NSRectFillList(rectList, n);
                }
              else
                {
                  NSRectFill(r[i]);
                }
            }

          if (l[i] > 0)
            {
              nswrap_moveshow(r[i].origin.x, r[i].origin.y,
                              t[i], l[i],
                              font,
                              (NS_FACE_FOREGROUND(face)!=nil ?
                               NS_FACE_FOREGROUND(face)
                               : FRAME_FOREGROUND_COLOR(nf)),
                              (flags!=NS_DUMPGLYPH_FOREGROUND ? nil
                               : (NS_FACE_BACKGROUND(face)!=nil ?
                                  NS_FACE_BACKGROUND(face)
                                  : FRAME_BACKGROUND_COLOR(nf))),
                              face->underline_p,
                              //PENDING: do in ns_frame_update_line_height
                              font->voffset + (NS_FRAME_LINE_HEIGHT(nf) -
                                               font->height)/2,
                              NS_FONT_WIDTH(nf));
            }
        }
    }

}

//23: folded into ns_dumpglyphs for now
void
nswrap_moveshow(float x, float y, const unsigned short txt[], int size,
                struct ns_font *font, NSColor *col, NSColor *bgCol,
                int ul, NSColor *ulColor, float voffset, float advances[],
                float twidth)
// ----------------------------------------------------------------------------
//  Move to target point and draw characters; voffset compensates for font var
// ----------------------------------------------------------------------------
{
    static char buf[1024];
    static float lastW = 0.0;
    float px, py;
    int i, pos;
    NSString *str;
    char *p;

    NSTRACE(nswrap_moveshow);

//NSLog(@"moveshow @ %.0f, %.0f: %d chars", x,y, size);

    px = x;
    py = y + voffset;

/*
    for (i=0, pos=0; i<size; i++)
      {
//          if (txt[i] & 0xff00)
        buf[pos++] = txt[i] >> 8;
        buf[pos++] = txt[i] & 0xff;
      }
    buf[pos] = 0;
*/
    for (i=0, p=buf; i<size; i++)
      CHAR_STRING_ADVANCE(txt[i], p);
    *p = 0;

#ifdef GNUSTEP
  // use DPS
  {
    NSGraphicsContext *ctxt = GSCurrentContext();

    // set up font and color
    DPSgsave(ctxt);
    [font->nsfont set];

    [col set];

    // do underline
    if (ul)
      {
        if (ulColor != nil) [ulColor set];
        DPSmoveto(ctxt, px, py+2);
        DPSlineto(ctxt, px+size*w, py+2);
        if (ulColor != nil) [col set];
      }

    // draw with DPSxshow()
    if (size > 1024)
      {
        fprintf(stderr, "nswrap_moveshow: run too long\n");
        return;
      }

    if (bgCol != nil)
      {
        [bgCol set];
        DPSmoveto(ctxt, px,py);
//[[NSGraphicsContext currentContext] GSSetTextDrawingMode: GSTextFillStroke]; /// not implemented yet
        DPSxshow(ctxt, buf, advances, size); // size is an upper bound..
        DPSstroke(ctxt);
        [col set];
//[[NSGraphicsContext currentContext] GSSetTextDrawingMode: GSTextFill]; /// not implemented yet
      }
    DPSmoveto(ctxt, px,py);
    DPSxshow(ctxt, buf, advances, size); // size is an upper bound..
    DPSstroke(ctxt);
    DPSgrestore(ctxt);
    return;
  }
#else

  str = [NSString stringWithUTF8String: buf];
//    str = [[NSString alloc] initWithBytes:buf length: size
//                                 encoding:NSUnicodeStringEncoding];
  if (!str) str = @".";

  // Use Quartz and an NSGlyphGenerator
  {
    static EmacsGlyphStorage *glyphStorage;
    static NSGlyphGenerator *glyphGenerator;
    static CGAffineTransform fliptf;
    static CGSize *sizes;
    static BOOL firstTime = YES;
    CGContextRef gcontext =
      [[NSGraphicsContext currentContext] graphicsPort];
    unsigned int gInd=0, cInd=0;

    if (firstTime == YES)
      {
        firstTime = NO;
        glyphStorage = [[EmacsGlyphStorage alloc] init];
        glyphGenerator = [NSGlyphGenerator sharedGlyphGenerator];
        fliptf = CGAffineTransformMakeScale(1.0,-1.0);
        sizes = (CGSize *) xmalloc(1024*sizeof(CGSize));
        bzero(sizes, 1024*sizeof(CGSize));
      }

    [glyphStorage setString: str font: font->nsfont];
    [glyphGenerator generateGlyphsForGlyphStorage: glyphStorage
                        desiredNumberOfCharacters: glyphStorage->maxChar
                                       glyphIndex: &gInd
                                   characterIndex: &cInd];

    CGContextSaveGState(gcontext); // doesn't hurt, but not needed
    [col set];

    if (ul)
      {
        if (ulColor != nil) [ulColor set];
        CGContextBeginPath(gcontext);
        CGContextMoveToPoint(gcontext, px, py+font->underpos);
        CGContextAddLineToPoint(gcontext, px+twidth, py+font->underpos);
        CGContextStrokePath(gcontext);
        if (ulColor != nil) [col set];
      }

    fliptf.c =  font->synthItal ? Fix2X(kATSItalicQDSkew) : 0.0;

    CGContextSetFont(gcontext, font->cgfont);
    CGContextSetFontSize(gcontext,font->size);

    if (ns_antialias_text == NO || font->size <= ns_antialias_threshold)
      CGContextSetShouldAntialias(gcontext, 0);
    else
      CGContextSetShouldAntialias(gcontext, 1);
    if (ns_use_qd_smoothing)
      CGContextSetFontRenderingMode(gcontext, 2); // 0 is Cocoa, 2 is QD

    CGContextSetTextMatrix(gcontext, fliptf);

    for (i=0; i<size; i++)
        sizes[i].width = advances[i];

    if (bgCol != nil)
      {
        // foreground drawing; erase first to avoid overstrike
        [bgCol set];
        CGContextSetTextDrawingMode(gcontext, kCGTextFillStroke);
        CGContextSetTextPosition(gcontext, px, py);
        CGContextShowGlyphsWithAdvances(gcontext,glyphStorage->cglyphs,
                                        sizes, glyphStorage->maxGlyph);
        CGContextSetTextDrawingMode(gcontext, kCGTextFill);
      }

    [col set];
    CGContextSetTextPosition(gcontext, px, py);
    CGContextShowGlyphsWithAdvances(gcontext,glyphStorage->cglyphs,
                                    sizes, glyphStorage->maxGlyph);

    CGContextRestoreGState(gcontext);
    return;
  }
#endif  // !GNUSTEP
}


//23: from ns_dumpcursor

#if 0
  pixel_to_glyph_coords(f, fx, fy, &nh, &nv, NULL, 1);
  //PENDING: clean up this code
  {
    struct glyph *current_glyphs = glyph_row->glyphs[TEXT_AREA];
    BOOL enabled = nh >= 0 && nv >= 0 && glyph_row->enabled_p;
    int max_h = nv >= 0 ? glyph_row->used[TEXT_AREA] : 0;
    int pad = (nh+1<max_h) && current_glyphs[nh+1].padding_p ? 1 : 0;
    struct face *face = FACE_FROM_ID (f, phys_cursor_glyph->face_id);
    XFontStruct *font = face->font;
    BOOL ital = font->synthItal ||
                 [[NSFontManager sharedFontManager] traitsOfFont: font->nsfont]
                     & NSItalicFontMask;
  }

  if (nh != -1)
    {
      lastH = nh;
      lastV = nv;
    }
#endif

    // fix up italic clipping for char 2 behind where we just moved to
//      if (ital && lastH-1 >= FIRST_CHAR_POSITION(f) && lastH==nh-1 && lastV==nv)
//        {
//          int cp = (current_glyphs[lastH-1].padding_p == 0)
//                       ? lastH-1 : lastH-2;
//          ns_dumpglyphs(f, cp, lastV, &(current_glyphs[cp]),
//                        1, NS_DUMPGLYPH_FOREGROUND);
//        }


//23: xdisp.c impl now suffices
void
pixel_to_glyph_coords(struct frame *f, int px, int py, int *x, int *y,
                      void *bounds /* ignored */, int noclip)
// ----------------------------------------------------------------------------
//   Convert from pixel coordinates to chars
// ----------------------------------------------------------------------------
{
  int fw = NS_FONT_WIDTH(f);
  int lh = NS_FRAME_LINE_HEIGHT(f);

  *x= (fw <= 0) ? 0 : floor(((float) px) / fw);
  *y= (lh <= 0) ? 0 : floor(((float) py) / lh);

  if (bounds)
    fprintf(stderr,"NS pixel_to_glyph_coord called with nonzero bounds\n");

  if (!noclip)
    {
      *x = IN_BOUND(0, *x, f->width);
      *y = IN_BOUND(0, *y, f->height);
    }
}

//23: xdisp.c impl now suffices
void
glyph_to_pixel_coords(struct frame *f,int x,int y,int *pix_x,int *pix_y)
// ----------------------------------------------------------------------------
//   Convert from char coordinates to pixel
// ----------------------------------------------------------------------------
{
  *pix_x=(int) rint(x * NS_FONT_WIDTH(f));
  *pix_y=(int) rint(y * NS_FRAME_LINE_HEIGHT(f));
}


//23: now uses x_write_glyphs in xdisp.c
static void ns_write_glyphs(GLYPH *start, int len)
{
  struct frame *f = (updating_frame ? updating_frame : selected_frame);

  TRACE(ns_write_glyphs);

  if (f != updating_frame)
    {
      curs_x = f->cursor_x;
      curs_y = f->cursor_y;
      ns_dumpcursor(11, f,f->cursor_x+len,f->cursor_y);
      ns_dumpglyphs(f, curs_x, curs_y, start, len,
                    highlight ? NS_DUMPGLYPH_HIGHLIGHT : NS_DUMPGLYPH_NORMAL);
    }
  else
    {
      if (curs_y == f->phys_cursor_y && curs_x <= f->phys_cursor_x
          && curs_x + len > f->phys_cursor_x)
        ns_dumpcursor(12, f,-1,-1);
      ns_dumpglyphs(f, curs_x, curs_y, start, len,
                    highlight ? NS_DUMPGLYPH_HIGHLIGHT : NS_DUMPGLYPH_NORMAL);
      curs_x += len;
    }

  ns_unfocus(f);
}


// 23: no longer needed
static void
ns_adjust_size(struct frame *f)
// ----------------------------------------------------------------------------
//   Adjust window pixel size to an even character grid size.
// ----------------------------------------------------------------------------
{
  int rows, cols, vbextra, fw, lh;
  NSView *view = FRAME_NS_VIEW(f);
  NSWindow *win = [view window];
  NSRect r;

  NSTRACE(ns_adjust_size);

  [win invalidateCursorRectsForView:view];

  // don't set up tracking rects as long as the window is deferred
  if ([win windowNumber] > 0)
    [(EmacsView *)view adjustTrackingRects];

  r = [view bounds];

  fw = NS_FONT_WIDTH(f);
  if (fw <= 0)
    cols = f->text_cols;
  else
    {
      vbextra = NS_SCROLL_BAR_WIDTH(f);
      f->output_data.ns->vertical_scroll_bar_extra = vbextra;
      cols = floor((r.size.width - vbextra) / fw + 0.5);
    }

  lh = FRAME_LINE_HEIGHT(f);
  if (lh <= 0)
    rows = f->text_lines;
  else
    rows = floor(r.size.height / lh + 0.5);

fprintf(stderr, "adjust_size: %d, %d\n",cols,rows);
  change_frame_size (f, rows, cols, 0, 1, 0);//pretend, delay, safe
  SET_FRAME_GARBAGED (f);
}


//23: simplified, since emacs now directs cursor rendering more closely
int dumpcursor_mask = (1<<3) /*| (1<<17) | (1<<16)*/;
void
ns_dumpcursor (int where, FRAME_PTR f, int nx, int ny)
// ----------------------------------------------------------------------------
//   Move the cursor to new indicated coordinates (erase and redraw)
//   Mask is devel tool to turn off cursor dumps from selected callers
// ----------------------------------------------------------------------------
{
  NSRect r,s;
  int pad;
  static int lastX = -1, lastY = -1;

  TRACE(dumpcursor);
  if (dumpcursor_mask & (1<<where))
    return;

  // Fast exit if everything is drawn OK
  if (f->phys_cursor_x == nx && f->phys_cursor_y == ny &&
      FRAME_CURSOR(f) == FRAME_NEW_CURSOR(f) &&
      [FRAME_CURSOR_COLOR(f) isEqual: FRAME_NEW_CURSOR_COLOR(f)])
    return;

  //PENDING: clean up this code
  {
    struct frame_glyphs *current_glyphs = FRAME_CURRENT_GLYPHS (f);
    int px = f->phys_cursor_x;
    int py = f->phys_cursor_y;
    BOOL enabled = nx >= 0 && ny >= 0 && current_glyphs->enable[ny];
    int max_x = ny >= 0 ? current_glyphs->used[ny] : 0;

    ///
    int faceid = FAST_GLYPH_FACE(f->phys_cursor_glyph);
    XFontStruct *font = faceid > -1 && faceid < FRAME_N_COMPUTED_FACES(f)
                     && FRAME_COMPUTED_FACE(f, faceid) != 0 ?
                 FRAME_COMPUTED_FACE(f, faceid)->font
                 : 0;
    BOOL ital = font ? font->synthItal ||
                 [[NSFontManager sharedFontManager] traitsOfFont: font->nsfont]
                     & NSItalicFontMask
                 : NO;

    // draw previous position(s) back to normal
    if (px >= FIRST_CHAR_POSITION(f) && py >= 0)
      {
        pad = px+1 < current_glyphs->used[py] &&
          current_glyphs->glyphs[py][px+1] & GLYPH_MASK_PADDING ? 1 : 0;
        ns_dumpglyphs(f, px, py, &(f->phys_cursor_glyph), 1+pad,
                      NS_DUMPGLYPH_NORMAL);
        // correct for synthetic italic clipping, case 1
        if (ital && px-1 >= FIRST_CHAR_POSITION(f) && px == nx-1 && py==ny)
          {
            int cp = (current_glyphs->glyphs[py][px-1] & GLYPH_MASK_PADDING)
              ? px-2 : px-1;
            ns_dumpglyphs(f, cp, py, &(current_glyphs->glyphs[py][cp]), 1,
                          NS_DUMPGLYPH_FOREGROUND);
          }
      }

    // correct for synthetic italic clipping, case 2
    if (ital && lastX-1 >= FIRST_CHAR_POSITION(f) && lastX == nx-1 && lastY==ny)
      {
        int cp = (current_glyphs->glyphs[lastY][lastX-1]
                  & GLYPH_MASK_PADDING) ? lastX-2 : lastX-1;
        ns_dumpglyphs(f, cp, lastY, &(current_glyphs->glyphs[lastY][cp]),
                      1, NS_DUMPGLYPH_FOREGROUND);
      }

    // update to new position
    if (nx != -1)
      {
        f->phys_cursor_glyph = ((enabled && (nx < max_x))
                                ? current_glyphs->glyphs[ny][nx]
                                : SPACEGLYPH);
        pad = (nx+1<max_x)
          && current_glyphs->glyphs[ny][nx+1] & GLYPH_MASK_PADDING ? 1 : 0;
      }
  }

  if (nx != -1)
    {
      lastX = nx;
      lastY = ny;
    }
  f->phys_cursor_x = nx;
  f->phys_cursor_y = ny;

  FRAME_CURSOR(f) = FRAME_NEW_CURSOR(f);
  FRAME_CURSOR_COLOR(f) = FRAME_NEW_CURSOR_COLOR(f);

  if (nx >= FIRST_CHAR_POSITION(f) && ny >= 0)
    {
      // draw the cursor
      ns_chars_to_rect(f, nx,ny, nx+pad,ny, &r);
      ns_focus(f, &r);
      [FRAME_BACKGROUND_COLOR(f) set];
      NSRectFill(r);
      [FRAME_CURSOR_COLOR(f) set];
      switch(FRAME_CURSOR(f))
        {
        case no_highlight:
          break;
        case filled_box:
          NSRectFill(r);
          break;
        case hollow_box:
          NSRectFill(r);
          [FRAME_BACKGROUND_COLOR(f) set];
          NSRectFill(NSInsetRect(r, 1, 1));
          [FRAME_CURSOR_COLOR(f) set];
          break;
        case bar:
          s=r;
          s.origin.y += 0.75 * s.size.height;
          s.size.height *= 0.25;
          NSRectFill(s);
          break;
        case line:
          s=r;
          s.size.width = 2;
          NSRectFill(s);
          break;
        }
      // draw the character under the cursor
      ns_dumpglyphs(f, nx, ny, &(f->phys_cursor_glyph), 1,
                    NS_DUMPGLYPH_CURSOR);
    }

  ns_unfocus(f);
}

// fragment from nswrap_moveshow for pre-10.3 OS X (can no longer easily be
// supported due to need to render text per specified advances, since emacs
// expects char width in string to be same as in isolation

  // Fallback approach that works on OS X and GNUstep, but suffers from
  // problem that NSLayoutManager tunes the spacing so that for certain
  // fonts and text characters, text goes off the grid that Emacs expects.
  {
    NSDictionary *attribs;
    NSNumber *num = nil;
    id objects[4], keys[4];
    int n = 0;

    if (font->nsfont)
      {
        keys[n] = (id) NSFontAttributeName;
        objects[n++] = font->nsfont;
      }
    if (col)
      {
        keys[n] = (id) NSForegroundColorAttributeName;
        objects[n++] = col;
      }
    if (ul)
      {
        num = [NSNumber numberWithInt: ul];
        keys[n] = (id) NSUnderlineStyleAttributeName;
        objects[n++] = num;
      }

    attribs = [NSDictionary dictionaryWithObjects: objects
                                          forKeys: keys
                                            count: n];

    [str drawAtPoint: NSMakePoint(px, py) withAttributes: attribs];
  }


///23: no longer used
static void ns_change_line_highlight(int new_highlight, int vpos,
                                     int first_unused_hpos)
{
  TRACE(ns_change_line_highlight);
  highlight = new_highlight;
  curs_x = 0;
  curs_y = vpos;
  ns_clear_end_of_line (updating_frame->width);
}


static void ns_clear_end_of_line(int first_unused)
{
  struct frame *f = (updating_frame ? updating_frame : selected_frame);
  BOOL left = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f);
  BOOL inMiniBuf = curs_y >= XFASTINT(XWINDOW(FRAME_MINIBUF_WINDOW(f))->top);
  BOOL needsCursor = NO;

  TRACE(ns_clear_end_of_line);
  // note, first_unused refers to never-used columns at right edge

  if (curs_y < 0 || curs_y >= f->height)
    return;
  if (first_unused <= 0)
    return;
  first_unused++;  // chop off italic overhang
  if (first_unused > f->width)
    first_unused = f->width;

  if (left)
    first_unused += 2 * FIRST_CHAR_POSITION(f);
  else if (inMiniBuf)
    first_unused += FRAME_SCROLL_BAR_COLS(f);

  if (curs_y == f->phys_cursor_y
      && curs_x <= f->phys_cursor_x && f->phys_cursor_x < first_unused)
    {
      ns_dumpcursor(7, f,-1,-1);
      needsCursor = YES;
    }

  ns_dumpglyphs(f, curs_x, curs_y,
                NULL, first_unused - curs_x,
                NS_DUMPGLYPH_BACKGROUND);


  if (needsCursor && f != updating_frame)
    ns_dumpcursor(8, f,f->cursor_x,f->cursor_y);
  ns_unfocus(f);
}


//23: this seems to have become disused in favor of XX_scroll_run
static void
ns_ins_del_lines(GLYPH *start, int len)
{
  TRACE(ns_ins_del_lines);
  abort();
}

//23: this was disused even in 20
static void
ns_delete_glyphs(int len)
{
  TRACE(ns_delete_glyphs);
  abort();
}

//23: this seems to have become disused in favor of other methods
static void
ns_set_terminal_window(int n)
{
  struct frame *f=(updating_frame ? updating_frame : selected_frame);
  TRACE(ns_set_terminal_window);
  flexlines = ((n > 0) && (n <= f->height)) ? n : f->height;
}

///23: no longer used
static void ns_reassert_line_highlight(int new, int vpos)
{
  TRACE(ns_reassert_line_highlight);
  highlight = new;
}


//23: from set_vertical_scrollbar

//    sb_left = WINDOW_HAS_VERTICAL_SCROLL_BAR_ON_LEFT (window) ?
//      left + width - sb_width : left;

  //PENDING: clear up junk next to scrollbar;
  // eventually solve by getting rid of empty space
//    if (view->scrollbarsNeedingUpdate > 0)
//        ns_clear_frame_area(f, left+sb_width, top, width-sb_width, height);

//...
  ns_chars_to_rect (f, left, top,
                    left + FRAME_SCROLL_BAR_COLS (f) - 1, top + height - 1, &r);
  if (FRAME_SCROLL_BAR_PIXEL_WIDTH (f) > 0)
    r.size.width = FRAME_SCROLL_BAR_PIXEL_WIDTH (f);

    // the parent view is flipped, so we need to flip y value
  v = [view frame];
  r.origin.y = (v.size.height + FRAME_INTERNAL_BORDER_WIDTH(f) -
                r.size.height - r.origin.y);
  // minor adjustment..
  if (FRAME_HAS_VERTICAL_SCROLL_BARS_ON_RIGHT (f))
    r.origin.x += 2*FRAME_INTERNAL_BORDER_WIDTH(f);

  if (v.size.height - (r.origin.y+r.size.height) < 5) // top scrollbar
    r.size.height += FRAME_INTERNAL_BORDER_WIDTH(f);


//23: from init area

/*
//23: Moved to redisplay_interface
  clear_end_of_line_hook = ns_clear_end_of_line;
  write_glyphs_hook = ns_write_glyphs;
  cursor_to_hook = ns_cursor_to;
//23: Dropped
  change_line_highlight_hook = ns_change_line_highlight;
  reassert_line_highlight_hook = ns_reassert_line_highlight;
*/


// ============================================================================
//
//    Mouse face now mostly done in xdisp.c (thank god)
//
// ============================================================================

/* This is used for debugging, to turn off note_mouse_highlight.  */
static int disable_mouse_highlight = 0;

static void
note_mouse_highlight (struct frame *f, int x, int y)
// ----------------------------------------------------------------------------
//   Take proper action when the mouse has moved to position X, Y on frame F
//   as regards highlighting characters that have mouse-face properties.
//   Also dehighlighting chars where the mouse was before.
// ----------------------------------------------------------------------------
{
  int row, column, portion;
  Lisp_Object window;
  struct window *w;
  int leftOffset = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f) ?
    FRAME_SCROLL_BAR_COLS(f) : 0;
  TRACE(note_mouse_highlight);

  if (disable_mouse_highlight)
    return;

  mouse_face_mouse_x = x;
  mouse_face_mouse_y = y;
  mouse_face_mouse_frame = f;

  /* Find out which glyph the mouse is on.  */
  pixel_to_glyph_coords (f, x, y, &column, &row, 0, 1);

  mouse_face_mouse_row = row;
  // set mouse_face_mouse_rect to cover current glyph for when we're
  // not in a mouse-face
  ns_chars_to_rect (f, column, row, column, row, &mouse_face_mouse_rect);
//fprintf(stderr, "mfm_rect 1: %d,%d + %d,%d went to %.0f,%.0f + %d, %d\n",
//        column,row,column,row, mouse_face_mouse_rect.origin.x,
//        mouse_face_mouse_rect.origin.y, (column-column+1), (row-row+1));

  if (mouse_face_defer)
    return;

  if (gc_in_progress)
    {
      mouse_face_deferred_gc = 1;
      return;
    }

  /* Which window is that in?  */
  window = window_from_coordinates (f, column, row, &portion);
  w = XWINDOW (window);

  w->left += leftOffset;

  /* If we were displaying active text in another window, clear that.  */
  if (! EQ (window, mouse_face_window))
    clear_mouse_face ();

    /* Are we in a window whose display is up to date?
       And verify the buffer's text has not changed.  */
  if (WINDOWP (window) && portion == 0
      && row >= 0 && column >= leftOffset
      && row < FRAME_HEIGHT (f) && column < FRAME_WIDTH (f)+leftOffset
      && EQ (w->window_end_valid, w->buffer)
      && w->last_modified == BUF_MODIFF (XBUFFER (w->buffer)))
    {
      int *ptr = FRAME_CURRENT_GLYPHS (f)->charstarts[row];
      int i, pos;

      /* Find which buffer position the mouse corresponds to.  */
      for (i = column; i >= leftOffset; i--)
        if (ptr[i] > leftOffset)
          break;
      pos = ptr[i];
      /* Is it outside the displayed active region (if any)?  */
      if (pos <= leftOffset)
        clear_mouse_face ();
      else if (! (EQ (window, mouse_face_window)
                  && row >= mouse_face_beg_row
                  && row <= mouse_face_end_row
                  && (row > mouse_face_beg_row
                      || column >= mouse_face_beg_col)
                  && (row < mouse_face_end_row
                      || column < mouse_face_end_col
                      || mouse_face_past_end)))
        {
          Lisp_Object mouse_face, overlay, position;
          Lisp_Object *overlay_vec;
          int len, noverlays;
          struct buffer *obuf;
          int obegv, ozv;

            /* If we get an out-of-range value, return now; avoid an error.  */
          if (pos > BUF_Z (XBUFFER (w->buffer)))
            {
              w->left -= leftOffset;
              return;
            }

          /* Make the window's buffer temporarily current for
               overlays_at and compute_char_face.  */
          obuf = current_buffer;
          current_buffer = XBUFFER (w->buffer);
          obegv = BEGV;
          ozv = ZV;
          BEGV = BEG;
          ZV = Z;

            /* Yes.  Clear the display of the old active region, if any.  */
          clear_mouse_face ();

          /* Is this char mouse-active?  */
          XSETINT (position, pos);

          len = 10;
          overlay_vec = (Lisp_Object *) xmalloc (len * sizeof (Lisp_Object));

          /* Put all the overlays we want in a vector in overlay_vec.
               Store the length in len.  */
          noverlays = overlays_at (XINT (pos), 1, &overlay_vec, &len,
                                   NULL, NULL);
          noverlays = sort_overlays (overlay_vec, noverlays, w);

            /* Find the highest priority overlay that has a mouse-face prop. */
          overlay = Qnil;
          for (i = 0; i < noverlays; i++)
            {
              mouse_face = Foverlay_get (overlay_vec[i], Qmouse_face);
              if (!NILP (mouse_face))
                {
                  overlay = overlay_vec[i];
                  break;
                }
            }
          free (overlay_vec);
          /* If no overlay applies, get a text property.  */
          if (NILP (overlay))
            mouse_face =
              Fget_text_property (position, Qmouse_face, w->buffer);

          /* Handle the overlay case.  */
          if (! NILP (overlay))
            {
              /* Find the range of text around this char that
                 should be active.  */
              Lisp_Object before, after;
              int ignore;

              before = Foverlay_start (overlay);
              after = Foverlay_end (overlay);
              /* Record this as the current active region.  */
              fast_find_position (window, before, ozv,
                                  &mouse_face_beg_col,
                                  &mouse_face_beg_row, leftOffset);
              mouse_face_past_end
                = !fast_find_position (window, after, ozv,
                                       &mouse_face_end_col,
                                       &mouse_face_end_row, leftOffset);
              mouse_face_window = window;
              mouse_face_face_id
                = compute_char_face (f, w, pos, 0, 0,
                                     &ignore, pos + 1, 1);

              /* Display it as active.  */
              show_mouse_face (1);
            }
          /* Handle the text property case.  */
          else if (! NILP (mouse_face))
            {
              /* Find the range of text around this char that
                 should be active.  */
              Lisp_Object before, after, beginning, end;
              int ignore;

              beginning = Fmarker_position (w->start);
              XSETINT (end, (BUF_Z (XBUFFER (w->buffer)) -
                             XFASTINT (w->window_end_pos)));

              before
                = Fprevious_single_property_change (make_number (pos + 1),
                                                    Qmouse_face,
                                                    w->buffer, beginning);
              after
                = Fnext_single_property_change (position, Qmouse_face,
                                                w->buffer, end);
              /* Record this as the current active region.  */
              fast_find_position (window, before, ozv,
                                  &mouse_face_beg_col,
                                  &mouse_face_beg_row, leftOffset);
              mouse_face_past_end
                = !fast_find_position (window, after, ozv,
                                       &mouse_face_end_col,
                                       &mouse_face_end_row, leftOffset);
              mouse_face_window = window;
              mouse_face_face_id
                = compute_char_face (f, w, pos, 0, 0,
                                     &ignore, pos + 1, 1);

              /* Display it as active.  */
              show_mouse_face (1);
            }
          /* Handle the no-mouse-face case */
          else
            {
              Lisp_Object tbefore, tafter, obefore, oafter;
              Lisp_Object beginning, end;
              int bcol, brow, ecol, erow;
              int width = window_internal_width (w) + w->left;
              int height = window_internal_height (w) + w->top;

              beginning = Fmarker_position (w->start);
              XSETINT (end, (BUF_Z (XBUFFER (w->buffer)) -
                             XFASTINT (w->window_end_pos)));

              if (pos+1 >= XFASTINT (end))
                {
                  fast_find_position (window, position, ozv,
                                      &bcol, &brow, leftOffset);
                  ecol = width;
                  erow = height;

                  if (bcol > column)
                    {
                      bcol = w->left;
                      brow++;
                    }
                }
              else
                {
                  int after, before;

                  tbefore =
                    Fprevious_single_property_change (make_number(pos+1),
                                                      Qmouse_face,
                                                      w->buffer,beginning);
                  tafter =
                    Fnext_single_property_change (position, Qmouse_face,
                                                  w->buffer, end);

                  obefore = Fprevious_overlay_change (make_number(pos+1));
                  oafter = Fnext_overlay_change (position);

                  if ((before = max (XFASTINT (tbefore),
                                     XFASTINT (obefore)))
                      == XFASTINT (beginning))
                    {
                      bcol = w->left;
                      brow = w->top;
                    }
                  else
                    {
                      fast_find_position (window, before, ozv,
                                          &bcol, &brow, leftOffset);

                      if (brow != row && bcol != (w->left+2))
                        {
                          brow++;
                          bcol = w->left;
                        }
                    }

                  if ((after = min (XFASTINT (tafter),
                                    XFASTINT (oafter))) == XFASTINT (end))
                    {
                      ecol = width;
                      erow = height;
                    }
                  else
                    {
                      fast_find_position (window, after, ozv,
                                          &ecol, &erow, leftOffset);

                      if (erow != row)
                        {
                          erow--;
                          ecol = width;
                        }
                      // we are in the row, but the region apparently
                      // ends before column.  This happens with
                      // invisible text, so we extend the region to the
                      // end of the line.
                      else if (ecol < column)
                        ecol = width;
                    }
                }
              ns_chars_to_rect (f, bcol, brow, ecol-1, erow,
                                &mouse_face_mouse_rect);
// fprintf(stderr, "mfm_rect 2: %d,%d + %d,%d went to %.0f,%.0f + %d, %d\n",
//         bcol,brow,ecol-1,erow, mouse_face_mouse_rect.origin.x,
//         mouse_face_mouse_rect.origin.y, (ecol-1-bcol+1), (erow-brow+1));
            }
          BEGV = obegv;
          ZV = ozv;
          current_buffer = obuf;
        }
    }
  w->left -= leftOffset;
}

/* Find the row and column of position POS in window WINDOW.
   Store them in *COLUMNP and *ROWP.
   This assumes display in WINDOW is up to date.
   If POS is above start of WINDOW, return coords
   of start of first screen line.
   If POS is after end of WINDOW, return coords of end of last screen line.

   Value is 1 if POS is in range, 0 if it was off screen.  */

static int
fast_find_position (Lisp_Object window,
                    int pos, int end, int *columnp, int *rowp,
                    int leftOffset)
{
  struct window *w = XWINDOW (window);
  FRAME_PTR f = XFRAME (WINDOW_FRAME (w));
  int i;
  int row = 0;
  int left = w->left;
  int top = w->top;
  int height = XFASTINT (w->height) - ! MINI_WINDOW_P (w);
  int width = window_internal_width (w);
  int *charstarts;
  int lastcol;
  int maybe_next_line = 0;

  TRACE(fast_find_position);
  /* Find the right row.  */
  for (i = 0;
       i < height;
       i++)
    {
      int linestart = FRAME_CURRENT_GLYPHS (f)->charstarts[top + i][left];
      if (linestart > pos)
        break;
      /* If the position sought is the end of the buffer,
         don't include the blank lines at the bottom of the window.  */
      if (linestart == pos && pos == end) // was BUF_ZV (XBUFFER(w->buffer)))
        {
          maybe_next_line = 1;
          break;
        }
      if (linestart > 0)
        row = i;
    }

  /* Find the right column with in it.  */
  charstarts = FRAME_CURRENT_GLYPHS (f)->charstarts[top + row];
  lastcol = left;
  for (i = leftOffset; i < width+leftOffset; i++)
    {
      if (charstarts[left + i] == pos)
        {
          *rowp = row + top;
          *columnp = i + left;
          return 1;
        }
      else if (charstarts[left + i] > pos)
        break;
      else if (charstarts[left + i] > leftOffset)
        lastcol = left + i;
    }

  /* If we're looking for the end of the buffer,
       and we didn't find it in the line we scanned,
       use the start of the following line.  */
  if (maybe_next_line)
    {
      row++;
      lastcol = left;
    }

  *rowp = row + top;
  *columnp = lastcol;
  return 0;
}


static void
show_mouse_face (int hl)
{
  struct window *w = XWINDOW (mouse_face_window);
  int width = window_internal_width (w);
  FRAME_PTR f = XFRAME (WINDOW_FRAME (w));
  int i;
  int cursor_off = 0;
  int old_curs_x = curs_x;
  int old_curs_y = curs_y;

  TRACE(show_mouse_face);

  /* Set these variables temporarily
       so that if we have to turn the cursor off and on again
       we will put it back at the same place.  */
  curs_x = f->phys_cursor_x;
  curs_y = f->phys_cursor_y;

  for (i = mouse_face_beg_row;
       i <= mouse_face_end_row; i++)
    {
      int column = (i == mouse_face_beg_row
                    ? mouse_face_beg_col : w->left);
      int endcolumn = (i == mouse_face_end_row
                       ? mouse_face_end_col : w->left + width);
      endcolumn = min (endcolumn, FRAME_CURRENT_GLYPHS (f)->used[i]);

      /* If the cursor's in the text we are about to rewrite,
         turn the cursor off.  */
      if (i == curs_y
          && curs_x >= column - 1
          && curs_x <= endcolumn)
        {
          ns_dumpcursor (4, f, -1, -1);
          cursor_off = 1;
        }

      ns_dumpglyphs (f, column, i,   // r.origin.x, r.origin.y,
                     FRAME_CURRENT_GLYPHS (f)->glyphs[i] + column,
                     endcolumn - column,
                     /* Highlight with mouse face if hl > 0.  */
                     hl > 0 ? NS_DUMPGLYPH_MOUSEFACE : NS_DUMPGLYPH_NORMAL);
      //PENDING: this was throwing off focus rect mgmt during updates, but
      // maybe it's OK now
      //ns_unfocus(f);

      if (i == mouse_face_mouse_row)
        {
          ns_chars_to_rect (f, column, i, endcolumn-1, i,
                            &mouse_face_mouse_rect);
        }

    }

  /* If we turned the cursor off, turn it back on.  */
  if (cursor_off)
      ns_dumpcursor(5, f,f->cursor_x,f->cursor_y);

  curs_x = old_curs_x;
  curs_y = old_curs_y;
  ns_unfocus(f);
}


static void
clear_mouse_face ()
{
  TRACE(clear_mouse_face);
  if (! NILP (mouse_face_window))
    {
      struct window *w = XWINDOW (mouse_face_window);
      int leftOffset
        = FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(mouse_face_mouse_frame)
        ? FRAME_SCROLL_BAR_COLS(mouse_face_mouse_frame) : 0;
      w->left += leftOffset;
      show_mouse_face (0);
      w->left -= leftOffset;
    }

  mouse_face_beg_row = mouse_face_beg_col = -1;
  mouse_face_end_row = mouse_face_end_col = -1;
  mouse_face_window = Qnil;
}


static void
ns_chars_to_rect (struct frame *f,int x1,int y1,int x2,int y2, NSRect *r)
// ----------------------------------------------------------------------------
//   Convert from char rectangle to AppKit rectangle
// ----------------------------------------------------------------------------
{
  int fw = NS_FONT_WIDTH(f);
  int lh = FRAME_LINE_HEIGHT(f);

  NSTRACE(ns_chars_to_rect);
  r->origin.x = x1 * fw;
  r->origin.y = y1 * lh;
  r->size.width = (x2 - x1 + 1) * fw;
  r->size.height = (y2 - y1 + 1) * lh;
}


static void
ns_rect_to_chars (struct frame *f, NSRect *r,
                  int *top, int *left, int *bottom, int *right)
// ----------------------------------------------------------------------------
//   Convert from AppKit rectangle to char rectangle
// ----------------------------------------------------------------------------
{
  int fw = NS_FONT_WIDTH(f);
  int lh = FRAME_LINE_HEIGHT(f);
  *left = r->origin.x / fw;
  *top  = r->origin.y / lh;
  *right = (r->origin.x + r->size.width + fw - 1)  / fw;
  *bottom = (r->origin.y + r->size.height + lh - 1)  / lh;
}



////////////////
// from nsterm.m but originally in nsfaces.m
////////////////



int ns_xlfd_to_fontsize(const char *xlfd)
// ----------------------------------------------------------------------------
// Pull font size out of an XLFD string, or 0 if not found.
// ----------------------------------------------------------------------------
{
  int s = 0;
  sscanf(xlfd, "-%*[^0-9]%d", &s);
fprintf(stderr, "scanned '%s' and got %d\n",xlfd, s);
  return s;
}


int
ns_frame_update_line_height (struct frame *f)
// ----------------------------------------------------------------------------
// Update the line_height of frame F according to the biggest font in
//  any face.  Return nonzero if if line_height changes.
// ----------------------------------------------------------------------------
{
  int i;
  int biggest = (int)rint(f->output_data.ns->font->height);
  struct face *face;

  NSTRACE(ns_frame_update_line_height);
//PENDING: update for new face data structures
//    for (i = 0; i < f->output_data.ns->n_param_faces; i++)
//      {
//        face = f->output_data.ns->param_faces[i];
//        if (face != 0)
//          {
//            if (face->font==FACE_DEFAULT/*0*/)
//              continue;
//            if ((face->font->height) > biggest)
//              biggest = (int)rint(face->font->height);
//          }
//      }

  if (biggest == FRAME_LINE_HEIGHT(f))
    return 0;

  FRAME_LINE_HEIGHT(f) = biggest;
fprintf(stderr, "setting line height to %d\n",biggest);
  return 1;
}


void
ns_frame_update_shrink_space(struct frame *f)
// ----------------------------------------------------------------------------
// NS-specific: Adjust param face font heights according to new shrinkspace.
// ----------------------------------------------------------------------------
{
  int i;
  float shrink, bbox3, linegap;
  struct face *face;
  XFontStruct *font;
  NSFont *sfont;
  NSRect rect;

  if (!NUMBERP(ns_shrink_space))
    error("No shrink space defined");
  shrink=QXFLOAT(ns_shrink_space);
//PENDING: update for new face data structures
#if 0
  for (i=-1; i<f->output_data.ns->n_param_faces; i++)
    {
      if (i == -1)
        font = f->output_data.ns->font;
      else
        {
          face = f->output_data.ns->param_faces[i];
          if (face == 0 || face->font == FACE_DEFAULT)
            continue;
          font = face->font;
        }
      sfont = [font->nsfont screenFont];
      if (sfont == nil)
        sfont = font->nsfont;
      rect = [sfont boundingRectForFont];
      bbox3 = rect.origin.y + rect.size.height;
      font->height=rint(shrink*(rect.size.height));
#if defined(COCOA) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3)
      font->voffset = shrink * bbox3;
#else
#if defined(GNUSTEP)
      font->voffset = (1-(1-shrink)/2.0) *
        (rect.size.height + [sfont descender]);
#else // OS X < 10.3
      linegap = [sfont defaultLineHeightForFont]
        - ([sfont ascender]-[sfont descender]);
      font->voffset =
        (fabs(linegap)<=0.7 || (linegap<0&&linegap>-0.9))
        ? ([sfont descender] - ((linegap<0 && linegap>-0.5)?4*linegap:0))
        : 0.0;
#endif // OS X < 10.3
#endif // ! OS X >= 10.3
    }
#endif
}


// fragment from ns_load_font (at end)
    if (f->output_data.ns->font == 0) {
        int fw;
        f->output_data.ns->font = font;
        fw = NS_FONT_WIDTH(f);
        if (f->config_scroll_bar_width > 0)
            f->config_scroll_bar_cols = (f->config_scroll_bar_width + fw-1) / fw;
        else
            f->config_scroll_bar_cols = (NS_SCROLL_BAR_WIDTH_DEFAULT + fw - 1) / fw;
        FRAME_LINE_HEIGHT(f) = FONT_HEIGHT (f->output_data.ns->font);
    }
// end fragment


// previously used to obtain bold fonts, etc.
DEFUN ("ns-convert-font-trait-internal",
       Fns_convert_font_trait_internal, Sns_convert_font_trait_internal,
       2, 2, 0, "")
     (Lisp_Object name, Lisp_Object attribute)
// ----------------------------------------------------------------------------
//  Used from ns-win.el: Access the font manager trait facility.
// ----------------------------------------------------------------------------
{
  id fm;
  id font;
  NSString *fname;
  BOOL synthItal = removeSynthItal(XSTRING(name)->data, &fname);

  NSTRACE(ns-convert-font-trait-internal);

  check_ns();
  CHECK_STRING(name);
  CHECK_SYMBOL(attribute);
  fm = [NSFontManager new];
  font = [NSFont fontWithName: fname size:10.0];
  if (!font)
    {
//fprintf(stderr, "Warning (convert-trait), font '%s' not found.\n",
//                XSTRING(name)->data);
      font = [NSFont userFixedPitchFontOfSize: 10.0];
    }
  if (!fm || !font)
    error("Unknown font.");

  if (EQ(attribute, intern ("italic")))
    {
      font=[fm convertFont:font toHaveTrait:NSItalicFontMask];
#if defined(COCOA) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3)
      if (!([fm traitsOfFont: font] & NSItalicFontMask))
        synthItal = YES;
#endif
    }
  else if (EQ(attribute, intern ("unitalic")))
    {
      font=[fm convertFont:font toNotHaveTrait:NSItalicFontMask];
#if defined(COCOA) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_3)
      synthItal = NO;
#endif
    }
  else if (EQ(attribute, intern ("bold")))
    font=[fm convertFont:font toHaveTrait:NSBoldFontMask];
  else if (EQ(attribute, intern ("unbold")))
    font=[fm convertFont:font toNotHaveTrait:NSBoldFontMask];
  else if (EQ(attribute, intern ("nonstandardcharset")))
    font=[fm convertFont:font toHaveTrait:NSNonStandardCharacterSetFontMask];
  else if (EQ(attribute, intern ("unnonstandardcharset")))
    font=[fm convertFont:font toNotHaveTrait:NSNonStandardCharacterSetFontMask];
  else if (EQ(attribute, intern ("narrow")))
    font=[fm convertFont:font toHaveTrait:NSNarrowFontMask];
  else if (EQ(attribute, intern ("unnarrow")))
    font=[fm convertFont:font toNotHaveTrait:NSNarrowFontMask];
  else if (EQ(attribute, intern ("expanded")))
    font=[fm convertFont:font toHaveTrait:NSExpandedFontMask];
  else if (EQ(attribute, intern ("unexpanded")))
    font=[fm convertFont:font toNotHaveTrait:NSExpandedFontMask];
  else if (EQ(attribute, intern ("condensed")))
    font=[fm convertFont:font toHaveTrait:NSCondensedFontMask];
  else if (EQ(attribute, intern ("uncondensed")))
    font=[fm convertFont:font toNotHaveTrait:NSCondensedFontMask];
  else if (EQ(attribute, intern ("smallcaps")))
    font=[fm convertFont:font toHaveTrait:NSSmallCapsFontMask];
  else if (EQ(attribute, intern ("unsmallcaps")))
    font=[fm convertFont:font toNotHaveTrait:NSSmallCapsFontMask];
  else if (EQ(attribute, intern ("poster")))
    font=[fm convertFont:font toHaveTrait:NSPosterFontMask];
  else if (EQ(attribute, intern ("unposter")))
    font=[fm convertFont:font toNotHaveTrait:NSPosterFontMask];
  else if (EQ(attribute, intern ("compressed")))
    font=[fm convertFont:font toHaveTrait:NSCompressedFontMask];
  else if (EQ(attribute, intern ("uncompressed")))
    font=[fm convertFont:font toNotHaveTrait:NSCompressedFontMask];
  else
    error("Unknown trait");

  fname = synthItal ? [[font fontName] stringByAppendingString: @"-synthItal"]
    : [font fontName];
  return build_string([fname UTF8String]);
}


////////////////
// from nsfns.m
////////////////

// Weird internal_border_width-related section from x-create-frame.
// Taking it out didn't seem to make any difference..
//   if (NILP (Fassq (Qinternal_border_width, parms)))
//     {
//       Lisp_Object value;

//       value = x_get_arg (dpyinfo, parms, Qinternal_border_width,
// 			 "internalBorder", "InternalBorder", RES_TYPE_NUMBER);
//       if (! EQ (value, Qunbound))
// 	parms = Fcons (Fcons (Qinternal_border_width, value),
// 		       parms);
//     }


// Apparently unnecessary now
DEFUN ("x-reset-menu",Fx_reset_menu, Sx_reset_menu, 0, 0, 0,
       "Cause the NS menu to be re-calculated.")
     ()
{
  set_frame_menubar(SELECTED_FRAME(),1, 0);
  return Qnil;
}
defsubr (&Sx_reset_menu);


///23: UNUSED
void ns_set_underline (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  store_frame_param (f, Qunderline, arg);
  if (FRAME_NS_VIEW(f))
      ns_recompute_basic_faces (f);

  SET_FRAME_GARBAGED (f);
}

///23: UNUSED
void ns_set_fontsize (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  Lisp_Object frame;

  if (XTYPE(arg)==Lisp_String)
    arg=Fstring_to_number(arg, Qnil);
  if (!NUMBERP(arg) || (XFLOATINT(arg)<=0.0))
    error ("Fontsize not a positive number.");

  store_frame_param (f, Qfontsize, arg);

  if (FRAME_NS_VIEW(f))
      ns_recompute_basic_faces (f);

  XSETFRAME (frame, f);
  call1 (Qface_set_after_frame_default, frame);
  SET_FRAME_GARBAGED (f);
}


// resource interface was reworked, and we no longer impl the lisp functions
DEFUN ("x-get-resource", Fx_get_resource, Sx_get_resource, 2, 4, 0,
       "Return the value of ATTRIBUTE, of class CLASS, from the X defaults database.\n\
This uses `INSTANCE.ATTRIBUTE' as the key and `Emacs.CLASS' as the\n\
class, where INSTANCE is the name under which Emacs was invoked, or\n\
the name specified by the `-name' or `-rn' command-line arguments.\n\
\n\
The optional arguments COMPONENT and SUBCLASS add to the key and the\n\
class, respectively.  You must specify both of them or neither.\n\
If you specify them, the key is `INSTANCE.COMPONENT.ATTRIBUTE'\n\
and the class is `Emacs.CLASS.SUBCLASS'.")
     (attribute, class, component, subclass)
     Lisp_Object attribute, class, component, subclass;
{
  register const char *value;

  check_ns ();

  CHECK_STRING (attribute, 0);
  CHECK_STRING (class, 0);

  if (!NILP (component))
    CHECK_STRING (component, 1);
  if (!NILP (subclass))
    CHECK_STRING (subclass, 2);
  if (NILP (component) != NILP (subclass))
    error ("x-get-resource: must specify both COMPONENT and SUBCLASS or neither");

  value=[[[NSUserDefaults standardUserDefaults]
               objectForKey:[NSString stringWithUTF8String:
                                       XSTRING (attribute)->data]] UTF8String];


  if (value != (char *) 0)
    return build_string (value);
//fprintf(stderr,"Nothing found for X resource '%s'.\n",XSTRING(attribute)->data);
  return Qnil;
}


//23: yet another function now genericized..
void x_set_frame_parameters (struct frame *f, Lisp_Object alist)
{
  Lisp_Object tail;

  /* If both of these parameters are present, it's more efficient to
     set them both at once.  So we wait until we've looked at the
     entire list before we set them.  */
  Lisp_Object width, height;

  /* Same here.  */
  Lisp_Object left, top;

  /* Same with these.  */
  Lisp_Object icon_left, icon_top;

  /* Record in these vectors all the parms specified.  */
  Lisp_Object *parms;
  Lisp_Object *values;
  int i;

  i = 0;
  for (tail = alist; CONSP (tail); tail = Fcdr (tail))
    i++;

  parms  = (Lisp_Object *) alloca (i * sizeof (Lisp_Object));
  values = (Lisp_Object *) alloca (i * sizeof (Lisp_Object));

  /* Extract parm names and values into those vectors.  */

  i = 0;
  for (tail = alist; CONSP (tail); tail = Fcdr (tail))
    {
      Lisp_Object elt;

      elt = Fcar (tail);
      parms[i] = Fcar (elt);
      values[i] = Fcdr (elt);
      //fprintf(stderr,"Setting '%s'\n", XSYMBOL(parms[i])->name->data);
      i++;
    }

  width = height = top = left = Qunbound;
  icon_left = icon_top = Qunbound;

  /* Now process them in reverse of specified order.  */
  for (i--; i >= 0; i--)
    {
      Lisp_Object prop, val;

      prop = parms[i];
      val = values[i];

      if (EQ (prop, Qwidth))
        width = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) :val;
      else if (EQ (prop, Qheight))
        height = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qtop))
        top = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qleft))
        left = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qicon_top))
        icon_top = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else if (EQ (prop, Qicon_left))
        icon_left = (XTYPE(val)==Lisp_String) ? Fstring_to_number(val, Qnil) : val;
      else
        {
          Lisp_Object param_index, old_value;

          param_index = Fget (prop, Qns_frame_parameter);
          old_value = get_frame_param (f, prop);

          store_frame_param (f, prop, val);
          if (XTYPE (param_index) == Lisp_Int
              && XINT (param_index) >= 0
              && (XINT (param_index)
                  < sizeof (ns_frame_parms)/sizeof (ns_frame_parms[0])))
            (*ns_frame_parms[XINT (param_index)].setter)(f, val, old_value);
        }
    }

  /* Don't die if just one of these was set.  */
  if (EQ (left, Qunbound))
    XSETINT (left, (int)f->output_data.ns->left);
  if (EQ (top, Qunbound))
    XSETINT (top, (int)f->output_data.ns->top);

  /* Don't die if just one of these was set.  */
  if (EQ (icon_left, Qunbound))
    icon_left = f->output_data.ns->icon_left;
  if (EQ (icon_top, Qunbound))
    icon_top = f->output_data.ns->icon_top;

  /* Don't die if just one of these was set.  */
  if (EQ (width, Qunbound))
    {
      if (FRAME_NEW_WIDTH (f))
        XSETINT (width, FRAME_NEW_WIDTH (f));
      else
        XSETINT (width, FRAME_WIDTH (f));
    }
  if (EQ (height, Qunbound))
    {
      if (FRAME_NEW_HEIGHT (f))
        XSETINT (height, FRAME_NEW_HEIGHT (f));
      else
        XSETINT (height, FRAME_HEIGHT (f));
    }

  /* Don't set these parameters these unless they've been explicitly
     specified.  The window might be mapped or resized while we're in
     this function, and we don't want to override that unless the lisp
     code has asked for it.

     Don't set these parameters unless they actually differ from the
     window's current parameters; the window may not actually exist
     yet.  */
  {
    Lisp_Object frame;

    XSETFRAME (frame, f);

    if (!INTEGERP (width) || !INTEGERP (height)
        || (INTEGERP (width) && XINT (width) != FRAME_WIDTH (f))
        || (INTEGERP (height) && XINT (height) != FRAME_HEIGHT (f))
        || FRAME_NEW_HEIGHT (f) || FRAME_NEW_WIDTH (f))
      Fset_frame_size (frame, width, height);
    if (!INTEGERP (left) || !INTEGERP (top)
        || (INTEGERP (left) && XINT (left) != f->output_data.ns->left)
        || (INTEGERP (top) && XINT (top) != f->output_data.ns->top))
      Fset_frame_position (frame, left, top);
    if ((! EQ (icon_left, f->output_data.ns->icon_left))
        || (! EQ (icon_top, f->output_data.ns->icon_top)))
      {
        NSWindow *window = [FRAME_NS_VIEW(f) window];
        NSScreen *screen = [window screen];

        f->output_data.ns->icon_top = icon_top;
        f->output_data.ns->icon_left = icon_left;
        if (NUMBERP (icon_top) && NUMBERP (icon_left))
          if ([window isMiniaturized])
            [window setFrameTopLeftPoint:
                      NSMakePoint(SCREENMAXBOUND(QXFLOAT (icon_left)),
                                  SCREENMAXBOUND([screen frame].size.height
                                  - QXFLOAT (icon_top)))];
      }
  }
}

//23: impl in frame.c.. but might need to augment for NS-specific stuff..
void x_report_frame_params (struct frame *f, Lisp_Object *alistptr)
{
  store_in_alist (alistptr, Qicon_left, f->output_data.ns->icon_left);
  store_in_alist (alistptr, Qicon_top, f->output_data.ns->icon_top);
  store_in_alist (alistptr, Qleft, make_number ((int) f->output_data.ns->left));
  store_in_alist (alistptr, Qtop, make_number ((int) f->output_data.ns->top));
  store_in_alist (alistptr, Qinternal_border_width,
                  make_number ((int) f->output_data.ns->internal_border_width));
  store_in_alist (alistptr, Qicon_name, f->icon_name);
  FRAME_SAMPLE_VISIBILITY (f);
  store_in_alist (alistptr, Qvisibility,
                  (FRAME_VISIBLE_P (f) ? Qt
                   : FRAME_ICONIFIED_P (f) ? Qicon : Qnil));
  store_in_alist (alistptr, Qdisplay,
                  XCONS (FRAME_NS_DISPLAY_INFO (f)->name_list_element)->car);
  store_in_alist (alistptr, Qmenu_bar_lines, FRAME_EXTERNAL_MENU_BAR(f));
}


DEFUN ("ns-show-ps", Fns_show_ps, Sns_show_ps,
       1, 1, 0, "Turn postscript logging on or off depending of the value of flag.")
     (flag)
     Lisp_Object flag;
{
  check_ns();
#if !defined(COCOA) && !defined(GNUSTEP)
  debug_log_dps_call("ns-show-ps: [NSDPSContext setAllContextsOutputTraced]");
  [NSDPSContext setAllContextsOutputTraced:!NILP(flag)];
#endif
  return Qnil;
}

DEFUN ("ns-show-events", Fns_show_events, Sns_show_events,
       1, 1, 0, "Turn event logging on or off depending of the value of flag.")
     (flag)
     Lisp_Object flag;
{
  check_ns();
#if !defined(COCOA) && !defined(GNUSTEP)
  debug_log_dps_call("ns-show-event: [NSDPSServerContext setEventsTraced]");
  [NSDPSServerContext setEventsTraced:!NILP(flag)];
#endif
  return Qnil;
}


DEFUN ("ns-utf-to-mule", Fns_utf_to_mule, Sns_utf_to_mule,
       1, 1, 0, "Convert given UTF-8 string to Mule string.")
     (str)
     Lisp_Object str;
{
  char *utfStr, *muleStr;
  int len;
  Lisp_Object retStr;

  CHECK_STRING (str);
  check_ns();

  utfStr = XSTRING(str)->data;
  len = strlen(utfStr);
  muleStr = (char *)xmalloc(len*3); // better estimate
  UTF8sToMulem(utfStr, len, muleStr);
  retStr = build_string(muleStr);
  xfree(muleStr);

  return retStr;
}

DEFUN ("ns-mule-to-utf", Fns_mule_to_utf, Sns_mule_to_utf,
       1, 1, 0, "Convert given Mule string to UTF-8 string.")
     (str)
     Lisp_Object str;
{
  char *muleStr, *utfStr;
  int len;
  Lisp_Object retStr;

  CHECK_STRING (str);
  check_ns();

  muleStr = XSTRING(str)->data;
  len = strlen(muleStr);
  utfStr = (char *)xmalloc(len*2); // better estimate
  mulemToUTF8s(muleStr, len, utfStr);
  retStr = build_string(utfStr);
  xfree(utfStr);

  return retStr;
}

static Lisp_Object ns_get_arg (Lisp_Object alist, Lisp_Object param,
                               char *owner, char *name)
{
  Lisp_Object tem;

  tem = Fassq (param, alist);
  if (!NILP(tem)) return Fcdr(tem);

  tem = Fassq (param, Vdefault_frame_alist);
  if (!NILP(tem)) return Fcdr(tem);

  if (!name) return Qunbound;

  tem = Fns_get_resource(owner ? build_string(owner) : Qnil,
                         build_string (name));

  if (NILP(tem)) return Qunbound;

  if (XTYPE(tem)==Lisp_String)
    {
      if (!strcmp(XSTRING(tem)->data,"YES")||
          !strcmp(XSTRING(tem)->data,"ON")||
          !strcmp(XSTRING(tem)->data,"TRUE")) return Qt;
      if (!strcmp(XSTRING(tem)->data,"NO")||
          !strcmp(XSTRING(tem)->data,"OFF")||
          !strcmp(XSTRING(tem)->data,"FALSE")) return Qnil;
    }

  return tem;
}

static Lisp_Object ns_default_parameter (struct frame *f, Lisp_Object alist,
                                         Lisp_Object prop, Lisp_Object deflt,
                                         char *owner, char *name,
                                         Lisp_Object *accum)
{
  Lisp_Object tem;

  tem = ns_get_arg (alist, prop, owner, name);
  if (EQ (tem, Qunbound)) tem=deflt;

  // accumulate into accum and caller is responsible for calling set()
  *accum = Fcons( Fcons (prop, tem), *accum);
  return tem;
}


void
ns_set_internal_border_width (struct frame *f, Lisp_Object arg,
                                   Lisp_Object oldval)
{
  if (XTYPE(arg)==Lisp_String)
    {
      arg=Fstring_to_number(arg, Qnil);
      store_frame_param(f,Qinternal_border_width,arg);
    }

  if (!NUMBERP(arg) || (XINT(arg)<0))
    {
      store_frame_param(f,Qinternal_border_width,oldval);
      error("Internal border width out of range");
    }

  f->output_data.ns->internal_border_width=XINT(arg);
  x_set_window_size (f, 0, f->text_cols, f->text_lines);
}


void
ns_set_visibility (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  Lisp_Object frame;
  XSETFRAME (frame, f);

  if (XTYPE(arg)==Lisp_String)
    {
      arg=Fintern(arg, Qnil);
      store_frame_param(f,Qvisibility,arg);
    }

  if (NILP (arg))
    Fmake_frame_invisible (frame, Qt);
  else if (EQ (arg, Qicon))
    Ficonify_frame (frame);
  else
    Fmake_frame_visible (frame);
}

void ns_set_autoraise (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  f->auto_raise = !NILP(arg);
}

void ns_set_autolower (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  f->auto_lower = !NILP(arg);
}

void ns_set_unsplittable (struct frame *f, Lisp_Object arg,
                          Lisp_Object oldval)
{
  f->no_split = !NILP (arg);
}


//23: old impl
void
ns_set_menu_bar_lines (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  FRAME_MENU_BAR_LINES (f) = 0;
  if (XTYPE(arg)==Lisp_String)
    {
      arg=Fstring_to_number(arg, Qnil);
    }

  if (FRAME_MINIBUF_ONLY_P(f))
    return;

  if (NILP(arg) || (XTYPE(arg)==Lisp_Int && XINT(arg)==0))
    {
      if (FRAME_EXTERNAL_MENU_BAR (f) == 1)
        free_frame_menubar (f);
      FRAME_EXTERNAL_MENU_BAR (f) = 0;
    }
  else
    {
      FRAME_EXTERNAL_MENU_BAR (f) = 1;
    }
}


void
ns_set_vertical_scroll_bars (struct frame *f, Lisp_Object arg,
                             Lisp_Object oldval)
{
  if ((EQ (arg, Qleft) && FRAME_HAS_VERTICAL_SCROLL_BARS_ON_RIGHT (f))
      || (EQ (arg, Qright) && FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT (f))
      || (NILP (arg) && FRAME_HAS_VERTICAL_SCROLL_BARS (f))
      || (!NILP (arg) && ! FRAME_HAS_VERTICAL_SCROLL_BARS (f)))
    {
      FRAME_VERTICAL_SCROLL_BAR_TYPE (f)
#if defined(COCOA)
        // Aqua guidelines mandate scroll bar on the right
        = (NILP (arg)
           ? vertical_scroll_bar_none
           : vertical_scroll_bar_right);
#else
      = (NILP (arg)
         ? vertical_scroll_bar_none
         : EQ (Qright, arg)
         ? vertical_scroll_bar_right
         : vertical_scroll_bar_left);
#endif  // COCOA

      /* We set this parameter before creating the X window for the
         frame, so we can get the geometry right from the start.
         However, if the window hasn't been created yet, we shouldn't
         call x_set_window_size.  */
      if (FRAME_NS_VIEW (f))
        x_set_window_size (f, 0, f->text_cols, f->text_lines);
    }
}


void
ns_set_scroll_bar_width (struct frame *f, Lisp_Object arg, Lisp_Object oldval)
{
  int wid = NS_FONT_WIDTH(f);
  int sbwid = NS_SCROLL_BAR_WIDTH_DEFAULT
#ifdef GNUSTEP
    - 1 // right side of sb is white, so ignore it
#endif
    ;
  if (!NILP(arg) && INTEGERP(arg) && XINT(arg) > 0)
    sbwid = XFASTINT(arg);

  FRAME_CONFIG_SCROLL_BAR_WIDTH (f) = sbwid;
  FRAME_CONFIG_SCROLL_BAR_COLS (f) = (sbwid + wid-1) / wid;
  if (FRAME_NS_VIEW (f))
    x_set_window_size (f, 0, f->text_cols, f->text_lines);
}
