/* NeXT/Open/GNUstep and MacOSX Cocoa menu and toolbar module.
   Copyright (C) 1986, 1988, 1993, 2005, 2006 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  

Originally by Carl Edman
Updated by Christian Limpach (chris@nice.ch)
OpenStep/Rhapsody  port by Scott Bender (sbender@harmony-ds.com)
MacOSX/Aqua port by Christophe de Dinechin (descubes@earthlink.net)
GNUstep port and post-20 update by Adrian Robert (arobert@cogsci.ucsd.edu)

*/

#include "config.h"
#include "lisp.h"
#include "window.h"
#include "commands.h"
#include "nsterm.h"
#include "termhooks.h"

//#include <sys/timeb.h>
//#include <sys/types.h>

#define MenuStagger 10.0

#if 0
int menu_trace_num = 0;
#define NSTRACE(x)        fprintf(stderr, "%s:%d: [%d] " #x "\n",        \
                                __FILE__, __LINE__, ++menu_trace_num)
#else
#define NSTRACE(x)
#endif

extern Lisp_Object Qundefined;
extern Lisp_Object Qmenu_enable;

Lisp_Object Vmenu_updating_frame;

Lisp_Object ns_menu_path;
EmacsMenu *mainMenu, *svcsMenu;

// NOTE: toolbar implementation is at end, following complete menu implementation


// ============================================================================
//
//    Menu: Externally-called functions
//
// ============================================================================


//23: PENDING: never called; may not be needed under NS?
void
x_activate_menubar (struct frame *f)
{
    fprintf(stderr, "Received x_activate_menubar event.\n");
}


void
free_frame_menubar (struct frame *f)
{
  EmacsMenu *menu = (EmacsMenu *) [NSApp mainMenu];

  NSTRACE(free_frame_menubar);

  [menu mark];
  [menu freeMarked];
}


/* Called to indicate that the frame's menus have changed, and the *step
   representation should be updated from Lisp.
   Performance note- this function gets called with every scroll.
   Should find way to detect and return from frivolous calls, but for now
   just need to make sure processing is efficient.  The EmacsMenu marking
   mechanism causes this to be a sub-10msec call after the first init for
   a window. */
void
set_frame_menubar (struct frame *f, int first_time, int deep_p)
{
  int i;
  Lisp_Object items,tail,key,string;
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
  BOOL needsSet = NO;
  id menu = [NSApp mainMenu];
  id attMenu;
//      struct timeb tb;
//      long t;

  NSTRACE(set_frame_menubar);
  //NSLog(@"got main menu: '%@'", menu);

  /** Menu may have been created automatically. **/
  if ([menu isKindOfClass:[EmacsMenu class]] == NO) 
    menu = nil;

  if (f != SELECTED_FRAME())
    {
      [pool release];
      return;
    }

  if (menu == nil)
    {
      menu = [[EmacsMenu alloc] initWithTitle: @"Emacs"];
      // set menu to not be autoupdated by NSApp.. must do ourselves
      [menu setAutoenablesItems: NO];
      needsSet = YES;
    }

  // close up anything on there
  attMenu = [menu attachedMenu];
  if (attMenu != nil)
    [attMenu close];

  // make sure we're playing with the latest deck
  //safe_run_hooks (Qmenu_bar_update_hook); // check if necessary
  if (NILP (items=FRAME_MENU_BAR_ITEMS (f)))
    {
      // no point, will soon be called again, just return
      //items= FRAME_MENU_BAR_ITEMS(f)=menu_bar_items(FRAME_MENU_BAR_ITEMS(f));
      [pool release];
      return;
    }
  items= FRAME_MENU_BAR_ITEMS(f)= menu_bar_items(FRAME_MENU_BAR_ITEMS(f));
//debug_print(items);
//ftime(&tb);
//t = -(1000*tb.time+tb.millitm);

  [menu mark];

  // this initializes/updates the Emacs menus from LISP
  for (i=0; i<XVECTOR(items)->size; i+=4)
    {
      key=XVECTOR(items)->contents[i];
      string=XVECTOR(items)->contents[i+1];
      tail=XVECTOR(items)->contents[i+2];
      if (NILP (string))
        break;

    //PENDING: below is correct, but apparently something isn't being
    // checked for in adjustKey that leads to a segfault down the road
    //  for (tail=XVECTOR(items)->contents[i+2]; CONSP(tail); tail=XCDR(tail)){
#ifdef COCOA
      if (!strcmp("Services",XSTRING(string)->data))
        {
          NSMenu *appMenu = [[mainMenu itemAtIndex: 0] submenu];
          NSMenuItem *cell =
            [appMenu itemAtIndex: [appMenu indexOfItemWithSubmenu:svcsMenu]];
          EmacsMenuObject *obj;
          if ([cell representedObject] == nil)
            {
              obj = [[[EmacsMenuObject alloc] init] autorelease];
              [obj setValue: key];
              [obj setSupercell: nil];
              [cell setRepresentedObject:obj];
            }
          [svcsMenu adjustMap: XCAR(tail)];
        }
      else
#endif
        [menu adjustKey:key binding:XCAR(tail) string:string];

      //PENDING: test whether we missed something (see PENDING above)
      for (tail=XCDR(tail); CONSP(tail); tail=XCDR(tail))
        fprintf(stderr, "ignoring top-level adjKey (please report as bug): '%s'\n",XSTRING(string)->data);

    }

//fprintf(stderr, "\nPRE free:\n");
//[menu dump: @"  "];

  [menu freeMarked];

//ftime(&tb);
//t += 1000*tb.time+tb.millitm;
//fprintf(stderr, "Menu update took %ld msec.\n",t);
//fprintf(stderr, "\nPOST free:\n");
//[menu dump: @"  "];

  // end init

  // set main menu
  if (needsSet)
    [NSApp setMainMenu: menu];

  [pool release];
}


Lisp_Object
ns_map_event_to_object (struct input_event *event, struct frame *f)
// ----------------------------------------------------------------------------
//    Called as part of a complicated chain between nsterm.m, keyboard.c,
//    and here, when a menu item is selected.  (If anyone wants to clean
//    this up, they should feel free to do so. ;)
// ----------------------------------------------------------------------------
{
  NSTRACE(ns_map_event_to_object);
  if (CONSP(ns_menu_path))
    {
      Lisp_Object ret = XCAR(ns_menu_path);
      ns_menu_path = XCDR(ns_menu_path);
      return ret;
    }
  return Qnil;
}



// ============================================================================
//
//    Menu: lisp definitions
//
// ============================================================================

DEFUN ("ns-reset-menu",Fns_reset_menu, Sns_reset_menu, 0, 0, 0,
       "Cause the NS menu to be re-calculated.")
     ()
{
  set_frame_menubar(SELECTED_FRAME(),1, 0);
  return Qnil;
}


DEFUN ("x-popup-menu",Fx_popup_menu, Sx_popup_menu, 2, 2, 0,
       "Pop up a deck-of-cards menu and return user's selection.\n\
POSITION is a position specification.  This is either a mouse button event\n\
or a list ((XOFFSET YOFFSET) WINDOW)\n\
where XOFFSET and YOFFSET are positions in pixels from the top left\n\
corner of WINDOW's frame.  (WINDOW may be a frame object instead of a window.)\n\
This controls the position of the center of the first line\n\
in the first pane of the menu, not the top left of the menu as a whole.\n\
\n\
MENU is a specifier for a menu.  For the simplest case, MENU is a keymap.\n\
The menu items come from key bindings that have a menu string as well as\n\
a definition; actually, the \"definition\" in such a key binding looks like\n\
\(STRING . REAL-DEFINITION).  To give the menu a title, put a string into\n\
the keymap as a top-level element.\n\n\
You can also use a list of keymaps as MENU.\n\
  Then each keymap makes a separate pane.\n\
When MENU is a keymap or a list of keymaps, the return value\n\
is a list of events.\n\n\
Alternatively, you can specify a menu of multiple panes\n\
  with a list of the form (TITLE PANE1 PANE2...),\n\
where each pane is a list of form (TITLE ITEM1 ITEM2...).\n\
Each ITEM is normally a cons cell (STRING . VALUE);\n\
but a string can appear as an item--that makes a nonselectable line\n\
in the menu.\n\
With this form of menu, the return value is VALUE from the chosen item.")
     (position, menu)
     Lisp_Object position, menu;
{
  // NOTE: this also gets called when a choice is made in the app main menu;
  //       the second part of ns-menu-path is sent back
  id panel;
  Lisp_Object window,x,y,tem;
  struct frame *f;
  NSPoint p;

  NSTRACE(x-popup-menu);

  if (NILP(position))
    {
      /* Don't create a menu.  Just precalculate the keyboard equivalents */
      return Qnil;
    }
  
  if (CONSP(ns_menu_path))
    {
      Lisp_Object ret = XCAR(ns_menu_path);
      ns_menu_path = XCDR(ns_menu_path);
      return ret;
    }
  
  check_ns();
  
  if (EQ (position, Qt))
    {
      /* Use the mouse's current position.  */
      struct frame *new_f = 0;

      if (mouse_position_hook)
        (*mouse_position_hook) (&new_f, 0, 0, 0, &x, &y, 0);
      if (new_f != 0)
        XSETFRAME (window, new_f);
      else
        {
          window = selected_window;
          x = make_number(0);
          y = make_number(0);
        }
    }
  else
    {
      CHECK_CONS (position);
      tem = Fcar (position);
      if (XTYPE (tem) == Lisp_Cons)
        {
          window = Fcar (Fcdr (position));
          x = Fcar(tem);
          y = Fcar(Fcdr (tem));
        }
      else
        {
          tem = Fcar (Fcdr (position));
          window = Fcar (tem);
          tem = Fcar (Fcdr (Fcdr (tem)));
          x = Fcar (tem);
          y = Fcdr (tem);
        }
    }
  
  CHECK_NUMBER (x);
  CHECK_NUMBER (y);
  
  if (FRAMEP (window))
    {
      f = XFRAME (window);
      
      p.x = 0;
      p.y = 0;
    }
  else
    {
      CHECK_LIVE_WINDOW (window);
      f = XFRAME (WINDOW_FRAME (XWINDOW (window)));
      
      p.x = ((int)(FRAME_COLUMN_WIDTH(f)) * XWINDOW (window)->left_col);
      p.y = (FRAME_LINE_HEIGHT(f) * XWINDOW (window)->top_line);
    }

  p.x += XINT(x); p.y += XINT(y);

  p = [FRAME_NS_VIEW(f) convertPoint: p toView: nil];
  p = [[FRAME_NS_VIEW(f) window] convertBaseToScreen: p];
  panel = [[EmacsMenuPanel alloc] initFromMenu: menu];
  
  tem = [panel runMenuAt: p.x : p.y];

  [panel close];

  [[FRAME_NS_VIEW(SELECTED_FRAME()) window] makeKeyWindow];
  return tem;
}


DEFUN ("x-popup-dialog", Fx_popup_dialog, Sx_popup_dialog, 2, 3, 0,
       doc: /* Pop up a dialog box and return user's selection.
POSITION specifies which frame to use.
This is normally a mouse button event or a window or frame.
If POSITION is t, it means to use the frame the mouse is on.
The dialog box appears in the middle of the specified frame.

CONTENTS specifies the alternatives to display in the dialog box.
It is a list of the form (DIALOG ITEM1 ITEM2...).
Each ITEM is a cons cell (STRING . VALUE).
The return value is VALUE from the chosen item.

An ITEM may also be just a string--that makes a nonselectable item.
An ITEM may also be nil--that means to put all preceding items
on the left of the dialog box and all following items on the right.
\(By default, approximately half appear on each side.)

If HEADER is non-nil, the frame title for the box is "Information",
otherwise it is "Question".

If the user gets rid of the dialog box without making a valid choice,
for instance using the window manager, then this produces a quit and
`x-popup-dialog' does not return.  */)
     (position, contents, header)
     Lisp_Object position, contents, header;
{
  id dialog;
  Lisp_Object window,tem;
  struct frame *f;
  NSPoint p;
  BOOL isQ;

  NSTRACE(x-popup-dialog);
  
  check_ns();

  isQ = NILP(header);

  if (EQ (position, Qt))
    {
      window = selected_window;
    }
  else if (CONSP (position))
    {
      Lisp_Object tem;
      tem = Fcar (position);
      if (XTYPE (tem) == Lisp_Cons)
        window = Fcar (Fcdr (position));
      else
        {
          tem = Fcar (Fcdr (position));  /* EVENT_START (position) */
          window = Fcar (tem);	     /* POSN_WINDOW (tem) */
        }
    }
  else if (FRAMEP (position))
    {
      window = position;
    }
  else
    {
      CHECK_LIVE_WINDOW (position);
      window = position;
    }
  
  if (FRAMEP (window))
    f = XFRAME (window);
  else
    {
      CHECK_LIVE_WINDOW (window);
      f = XFRAME (WINDOW_FRAME (XWINDOW (window)));
    }
  p.x = (int)f->left_pos + ((int)FRAME_COLUMN_WIDTH(f) * f->text_cols)/2;
  p.y = (int)f->top_pos + (FRAME_LINE_HEIGHT(f) * f->text_lines)/2;
  dialog=[[EmacsDialogPanel alloc] initFromContents: contents isQuestion: isQ];
  
  tem=[dialog runDialogAt: p.x : p.y];

  [dialog close];

  [[FRAME_NS_VIEW(SELECTED_FRAME()) window] makeKeyWindow];
  return tem;
}


// ============================================================================
//
//    Menu: lisp interface declaration
//
// ============================================================================

void
syms_of_nsmenu ()
{
  ns_menu_path=Qnil;
  staticpro(&ns_menu_path);
  defsubr (&Sx_popup_menu);
  defsubr (&Sx_popup_dialog);
  defsubr (&Sns_reset_menu);
  DEFVAR_LISP ("menu-updating-frame", &Vmenu_updating_frame,
    "Frame for which we are updating a menu.\n\
The enable predicate for a menu command should check this variable.");
  Vmenu_updating_frame = Qnil;
}



// ============================================================================
//
//    Menu: class implementations
//
// ============================================================================


/* Link btwn NS menuitem and keymap lisp command object. */
@implementation EmacsMenuObject

- init
{
  [super init];
  value = Qnil;
  supercell = nil;
  return self;
}


- (void)dealloc
{
  [super dealloc]; 
  return;
}


- (Lisp_Object) value
{
  return value;
}


- setValue:(Lisp_Object)nvalue
{
  // note, care must be taken not to retain/use this ref past when lisp
  // garbage-collects it (and it cannot easily be protected from such)
  if (nvalue != value)
    value=nvalue;
  return self;
}


- supercell
{
  return supercell;
}


- setSupercell:(id) nsupercell
{
  supercell=nsupercell;
  return self;
}

@end


/*
   Implementation for menubar menus, with machinery to optimize updating.
   When an update is initiated, '-mark' is called, which sets a marker in
   menu and all submenus to 0.  As items are parsed from lisp they are
   checked to see if they are the same as existing items.  If so, the
   existing item is kept, otherwise a new replacement item is inserted
   at the marker position.  In either case, the marker is advanced.  Once
   the round is complete, -freeMarked is called, which deletes all items
   BEYOND the marker, which are old items that were not reused.
 */
@implementation EmacsMenu

- init
{
  [super init];
  minmarked=0;
  return self;
}


- (BOOL)performKeyEquivalent:(NSEvent *)theEvent 
{
  if (SELECTED_FRAME() && FRAME_NS_VIEW(SELECTED_FRAME()))
    [FRAME_NS_VIEW(SELECTED_FRAME()) keyDown:theEvent];
  return YES;
}


// inserts item at position 'minmarked' (-addMinItem...)
- addItem: (Lisp_Object) val
   string: (Lisp_Object) aString
     bind: (Lisp_Object)bind
{
  unsigned int key = 0;
  id cell;
  NSString *title, *keyEq;
  EmacsMenuObject *obj;

  if (!NILP(bind))
    {
      Lisp_Object tail,vec;
      for (tail=Fwhere_is_internal(bind,Qnil,Qnil,Qt);
           XTYPE(tail)==Lisp_Cons;
           tail=XCDR(tail))
        {
          vec = XCAR(tail);
          if (!VECTORP(vec) ||
              XVECTOR(vec)->size != 1 ||
              XTYPE(XVECTOR(vec)->contents[0]) != Lisp_Int)
            continue;
          key = XFASTINT(XVECTOR(vec)->contents[0]);
          if ((key&(alt_modifier|hyper_modifier|super_modifier))
              ==super_modifier)
            {
              key &= 0xff;
              break;
            }
          else
            {
              key = 0;
            }
        }
    }

  title = [NSString stringWithUTF8String: XSTRING(aString)->data];
  if (title == nil)
      title = @"< ? >";

  keyEq = key == 0 ? @"" : [NSString stringWithFormat:@"%c", key];
//fprintf(stderr, "Adding menuItem '%s' w/keyEq '%s', val is of type %d\n", XSTRING(aString)->data, [keyEq cString], XTYPE(val));
//debug_print(val);
  cell = [self addMinItemWithTitle: title
                            action: @selector(menuDown:)
                     keyEquivalent: keyEq];
  ///[cell setEnabled: YES];

  obj = [[[EmacsMenuObject alloc] init] autorelease];
  [obj setValue:val];
  [obj setSupercell:supercell];
  
  [cell setRepresentedObject:obj];
  
  return cell;
}


- supercell
{
  return supercell;
}


- setSupercell:(id) nsupercell
{
  supercell=nsupercell;
  return self;
}


- mark
{
  NSEnumerator *ienum;
  id <NSMenuItem> item;

  // strictly only needed on GNUstep/Art, but is correct style
  [self setMenuChangedMessagesEnabled: NO];

  ienum = [[self itemArray] objectEnumerator];
  while ( (item = [ienum nextObject]) )
    if ([item hasSubmenu])
      [(EmacsMenu *)[item submenu] mark];
#ifdef COCOA
  // this protects the app menu
  if ([self supermenu]==nil && ![[self title] isEqualToString: @"Services"])
    minmarked = 1;
  else if ([[self title] length] == 0 ||
           [[self title] isEqualToString: @"Apple"])
    minmarked = [[self itemArray] count];
  else
#endif
    minmarked = 0;

  return self;
}


- freeMarked
{
  int i;
  id cell;
  NSArray *items = [self itemArray];
  int nFreed = [items count] - minmarked;
//if (nFreed > 0) fprintf(stderr, "FreeMarked: %s: freeing %d items\n", [[self title] cString], nFreed);

  i = [items count];
  while (i > minmarked && ((cell=[items objectAtIndex:--i])!=nil))
    [self removeItem:cell];

  for (i=0;i < [items count];i++)
    {
      cell = [items objectAtIndex:i];
      if ([cell hasSubmenu])
        [(EmacsMenu *)[cell submenu] freeMarked];
    }

  // see note above in -mark
  [self setMenuChangedMessagesEnabled: YES];

  //PENDING: track whether anything was added too (seems auto on OS X)
#ifdef COCOA
  if ([self supermenu] == nil && (nFreed > 0))
    [self sizeToFit];
#else
  if ([[self window] isVisible])
    [self sizeToFit];
#endif
  return self;
}


- findCell:(Lisp_Object)value
{
  NSEnumerator *ienum;
  id <NSMenuItem> item;
  
  ienum = [[self itemArray] objectEnumerator];
  while ( (item = [ienum nextObject]) )
    if ( [(EmacsMenuObject *)[item representedObject] value] == value )
      return item;
  
  return nil;
}


- (int)findCellNumber:(Lisp_Object)value
{
  id cell;
  NSArray *items = [self itemArray];
  int i;
  
  for (i=0; i < [items count]; i++)
    {
      cell = [items objectAtIndex: i];
      if ([(EmacsMenuObject *)[cell representedObject] value] == value) 
        return i;
    }
  return -1;
}

- (id <NSMenuItem>)addMinItemWithTitle: (NSString *)aString 
                                action: (SEL)aSelector 
                         keyEquivalent: (NSString *)charCode
{
  return [self insertItemWithTitle: aString 
                            action: aSelector 
                     keyEquivalent: charCode 
                           atIndex: minmarked];  
}


- removeCell:cell
{
  int i = [self findCellNumber:
                  [(EmacsMenuObject *)[cell representedObject] value]];
  
  if (i != -1)
    {
      [self removeItem: cell];
    
      if (i < minmarked) 
        minmarked--;
    }
  return self;
}


- adjustMap: (Lisp_Object)keymap;
{
  Lisp_Object item;

  for (keymap=XCDR(get_keymap(keymap));
       XTYPE(keymap) == Lisp_Cons;
       keymap = XCDR(keymap))
    {
      item = XCAR(keymap);

      if (XTYPE(item) == Lisp_Cons)
        {
          [self adjustKey: XCAR(item) binding: XCDR(item)];
        }
      else if (VECTORP(item))
        {
          int len = XVECTOR (item)->size,c;
          for (c=0; c<len; c++)
            [self adjustKey: c binding: XVECTOR(item)->contents[c]];
        }
      // (otherwise item was a string or a separator)
    }
  return self;
}


// called internally, for submenus and items
- (void)adjustKey:(Lisp_Object)key binding:(Lisp_Object)bind
{
  if (XTYPE(bind) == Lisp_Cons)
    [self adjustKey: key binding: XCDR(bind) string: XCAR(bind)];
  else if (EQ (bind, Qundefined))
    [self removeCell:[self findCell:key]];
  return;
}


// called externally, and also internally, to create new menuitems
- (void)adjustKey:(Lisp_Object)key
          binding:(Lisp_Object)bind
           string:(Lisp_Object)string
{
  id cell = nil;
  NSString *str;

  // this path occurs for menus created w/easymenu.el
  if (XTYPE(string) != Lisp_String)
    {
//fprintf(stderr, "Nonstring; symbol name=%d, binding of type %d, car/cdr = %s / %d\n",(XSYMBOL(string)->name)->data, XTYPE(bind), XSTRING(XCAR(bind))->data,XTYPE(XCDR(bind)));
      if (XTYPE(bind) == Lisp_Cons && !NILP(XCAR(bind)))
        {
          if (CONSP(XCDR(bind)))
            {
              [self adjustKey:key binding:XCAR(XCDR(bind))
                       string:XCAR(bind)];
            }
          else
            {
              [self adjustKey:key binding:XCDR(bind)
                       string:XCAR(bind)];
            }
        }
      return;
    }

  if ([[self itemArray] count] > minmarked
      && ((cell = [[self itemArray] objectAtIndex: minmarked]) != nil) 
      && (([cell isSeparatorItem] &&(!strlen(XSTRING(string)->data) ||
                                     !strncmp("--",XSTRING(string)->data,2)))
          ||
          (([(EmacsMenuObject *)[cell representedObject] value] == key)
           && !strcmp([[cell title] UTF8String], XSTRING(string)->data)
           && [cell isEnabled])))
    {

      /* Keep item, but check (below) that its subitems are the same. */
      if ([cell isSeparatorItem])
        {
          minmarked++;
          return;
        }

    }
  else if (!NILP(Fkeymapp(bind)))
    {
      // new submenu
      cell = [self addItem:key string:string bind:Qnil];
    }
  else
    {
      // not sure what this case is, but xmenu.c also drops it
      if (!(SYMBOLP(bind) || (CONSP(bind) && NILP(Fkeymapp(bind)))))
        return;
    // new single item
      if (!strlen(XSTRING(string)->data) ||
          !strncmp("--",XSTRING(string)->data,2))
        {
          [self insertItem: [NSMenuItem separatorItem] atIndex: minmarked++];
          return;
        }
      else
        {
          cell = [self addItem:key string:string bind:bind];
        }
    }

  str = [NSString stringWithUTF8String: XSTRING(string)->data];

  if (!NILP(Fkeymapp(bind)))
    {
      id nmenu;
     
      if (![cell hasSubmenu])
        {
          nmenu = [[EmacsMenu alloc] initWithTitle: str];
          [nmenu setAutoenablesItems: NO];
          [nmenu setSupercell:cell];         
          [self setSubmenu:nmenu forItem:cell];
          [nmenu release];
        }
      else
        {
          nmenu = [cell submenu];
        }
     
      [nmenu mark];
      [nmenu adjustMap:bind];
    }
  else
    {
      if ([cell hasSubmenu])
        {
          [self removeCell: cell];
          cell = [self addItem:key string:string bind:bind];
        }
    }

  if (![[cell title] isEqualToString: str])
    [cell setTitle:str];

  //PENDING: this is too slow to be done this regularly; menu mgmt logic
  //         will eventually need to be changed to parallel the ugly xmenu.c
//    parse_menu_item(Fcons(string,bind),0,0);
//    if (!NILP(XVECTOR(item_properties)->contents[ITEM_PROPERTY_TYPE]))
//        fprintf(stderr, "Found a toggle item: %s\n", XSTRING(string)->data);
//    [cell setEnabled: 
//             NILP(XVECTOR(item_properties)->contents[ITEM_PROPERTY_ENABLE]) ?
//          NO : YES];
  [cell setEnabled: YES];

  minmarked++;
}


-(void)setSubmenu: (NSMenu *)menu forItem: (id<NSMenuItem>)item
{
  if (![menu isKindOfClass: [EmacsMenu class]])
    {
      // this occurs unavoidably for the services menu; limit the damage
//fprintf(stderr, "not an emacs menu: '%s'\n", [[menu title] cString]);
//[NSException raise: @"NotAnEmacsMenu" format: @"Not an EmacsMenu. "];
      [menu setAutoenablesItems: NO];
      [menu retain];
    }
  else
    {
      [super setSubmenu: menu forItem: item];
    }
}


- dump: (NSString *)indent
{
  NSEnumerator *ienum;
  id <NSMenuItem> item;
  NSString *subindent = [indent stringByAppendingString: @"    "];
  int pos = 0;

  ienum = [[self itemArray] objectEnumerator];
  while ( (item = [ienum nextObject]) )
    {
      BOOL hasSub = [item hasSubmenu];
      if (pos++ == minmarked)
        fprintf(stderr, "%sMARK\n", [indent cString]);
      fprintf(stderr, "%s%s%s\n",[indent cString],[[item title] UTF8String],
              (hasSub ? " >" : ""));
      if (hasSub)
        {
          NSMenu *sub = [item submenu];
          if (![sub isKindOfClass: [EmacsMenu class]])
            fprintf(stderr, "%s(not an EmacsMenu)\n", [subindent cString]);
          else
            [(EmacsMenu *)sub dump: subindent];
        }
    }
  return self;
}

@end


/* One menu panel in a popup menu. */
@implementation EmacsMenuCard

- init
{
  [super init];
  [self setFont:[NSFont userFixedPitchFontOfSize:0]];
  [self setTitle:@""];
  [self setPullsDown:YES];
  [self setTarget:self];
  [self setAction:@selector(clicked:)];
  [self setAlignment:NSCenterTextAlignment];
#ifdef COCOA
  //PENDING: can now select from whole menu, but don't see "Select"
  [self addItem: "Select..." value: 0 enabled: YES];
#endif
  return self;
}


- addTitle: (char *)name
{
  NSString *title;

  title = [NSString stringWithUTF8String: name];
  if (title == nil)
    title = @"< ? >";

  [self setTitle: title];
  return self;
}


- addItem: (char *)name value:(Lisp_Object)val enabled:(BOOL)enabled
{
  NSRect r;
  id item;
  float f;
  NSString *title;

  title = [NSString stringWithUTF8String: name];
  if (title == nil)
      title = @"< ? >";

  [self addItemWithTitle: title];
  item = [self itemAtIndex:[self numberOfItems]-1];
  [item setTag:(int)val];

  if (!enabled)
      [item setEnabled:NO];

  //PENDING: make sure enough room for title (+ buttons) too
  f = [[self font] widthOfString: [item title]] + 12.0;
  r = [self frame];
  if (f > r.size.width)
    [self setFrameSize: NSMakeSize(f, r.size.height)];

  return self;
}


- (void)clicked:sender
{
  Lisp_Object seltag;
  id selected;
  
  if ( (selected = [self selectedItem]) && [selected tag] )
    {
      seltag = (Lisp_Object)[selected tag];
      if (! EQ(seltag,Qundefined))
        [NSApp stopModalWithCode:seltag];
    }
}
@end



/* A popup menu. */
@implementation EmacsMenuPanel
- initWithContentRect: (NSRect)contentRect 
            styleMask: (unsigned int)aStyle 
              backing: (NSBackingStoreType)backingType 
                defer: (BOOL)flag
{
  aStyle=NSTitledWindowMask;
  flag=YES;
  list=nil;
  maxwidth=0;
  [super initWithContentRect:contentRect 
                   styleMask:aStyle|NSClosableWindowMask 
                     backing:backingType 
                       defer:flag];
  [self setOneShot:YES];
  [self setReleasedWhenClosed:YES];
  [self setHidesOnDeactivate:YES];
  return self;
}


- (BOOL)windowShouldClose:(id)sender
{
  [NSApp stopModalWithCode:Qnil];
  return NO;
}


static void process_keymap(id panel, Lisp_Object keymap, Lisp_Object prefix,
                           char *name)
{
  Lisp_Object tail,pending_maps,tem;
  int card;
  
  pending_maps=Qnil;
  
  card = [panel cards];
  if (name)
    [panel addTitle: name to: card];
  else
    [panel addTitle: "Select..." to: card];

  for (tail = get_keymap(keymap); XTYPE(tail) == Lisp_Cons;
       tail = XCDR(tail))
    {
      Lisp_Object item = XCAR(tail), item1, item2, def, enabled;
      if (XTYPE(item) == Lisp_Cons)
        {
          item1 = XCDR(item);
          if (XTYPE(item1) == Lisp_Cons)
            {
              item2 = XCAR(item1);
              if (XTYPE(item2) == Lisp_String)
                {
                  def = XCDR(item1);
                  enabled = Qt;
                  if (XTYPE(def) == Lisp_Symbol)
                    {
                      tem = Fget(def, Qmenu_enable);
                      if (!NILP(tem))
                        enabled = Feval(tem);
                    }
                  if (XSTRING (item2)->data[0] == '@' && !NILP(Fkeymapp(def)))
                    pending_maps = Fcons(Fcons(def,
                                               Fcons(item2, XCAR(item))),
                                         pending_maps);
                  else
                    [panel addItem: XSTRING(item2)->data
                             value: Freverse(Fcons(XCAR(item), prefix))
                           enabled: !NILP(enabled)
                                at: card];
                }
            }
        }
      else if (VECTORP(item))
        {
          int len = XVECTOR(item)->size;
          int c;
          for (c = 0; c < len; c++)
            {
              Lisp_Object character;
              XSETFASTINT(character, c);
              item1 = XVECTOR(item)->contents[c];
              if (XTYPE(item1) == Lisp_Cons)
                {
                  item2 = XCAR(item1);
                  if (XTYPE(item2) == Lisp_String)
                    {
                      def = XCDR(item1);
                      enabled = Qt;
                      if (XTYPE(def) == Lisp_Symbol)
                        {
                          tem = Fget(def,Qmenu_enable);
                          if (!NILP(tem))
                            enabled = Feval(tem);
                        }
                      if (XSTRING (item2)->data[0] == '@' &&
                          !NILP(Fkeymapp(def)))
                        pending_maps = Fcons(Fcons(def,
                                                   Fcons(item2,
                                                         XCAR(item))),
                                             pending_maps);
                      else
                        [panel addItem: XSTRING(item2)->data
                                 value: Freverse(Fcons(XCAR(item),prefix))
                               enabled: !NILP(enabled)
                                    at:card];
                    }
                }
            }
        }
    }
  
  for (tail = pending_maps; XTYPE(tail) == Lisp_Cons; tail = XCDR(tail))
    {
      tem = XCAR(tail);
      process_keymap(panel,
                     XCAR(tem),
                     Fcons(XCDR(XCDR(tem)), prefix),
                     (char *)XSTRING(XCAR(XCDR(tem)))->data + 1);
    }
}


void process_list(id panel, Lisp_Object list)
{
  Lisp_Object tail, pane, tem;
  int card;
  
  card = [panel cards];
  for ( ; XTYPE(list) == Lisp_Cons; list = XCDR(list))
    {
      pane = XCAR(list);
      if (XTYPE(pane) != Lisp_Cons || XTYPE(XCAR(pane)) != Lisp_String)
        continue;
      for (tail = pane; XTYPE(tail) == Lisp_Cons; tail = XCDR(tail))
        {
          tem = XCAR(tail);
          if (XTYPE(tem) == Lisp_String)
            {
              if (tail != pane)		/* Not very first item? */
                [panel addItem: XSTRING(tem)->data
                         value: Qnil
                       enabled: NO
                            at: card];
              else
                {
                  char *s = XSTRING(tem)->data;
                  if (!s || strlen(s) == 0 || (strlen(s) == 1 && s[0] == ' '))
                    s = "Select...";
                  [panel addTitle: s to: card];
                }
            }
          else if (XTYPE(tem) == Lisp_Cons ||
                   XTYPE(XCAR(tem)) == Lisp_String)
            {
              [panel addItem: XSTRING(XCAR(tem))->data
                       value: XCDR(tem)
                     enabled:YES
                          at: card];
            }
        }
    }
}


- addItem:(char *)str value:(Lisp_Object)val enabled:(BOOL)enabled at:(int)pane
{
  int row;
  EmacsMenuCard *card;
  NSRect area;
  
  for (row = [list count]; row <= pane; row++)
    {
      card = [[EmacsMenuCard alloc] init];
      [list addObject: card];
      [[self contentView] addSubview:card];

      area.size = [[card cell] cellSize];
      if (area.size.width > maxwidth)
        maxwidth = area.size.width;
      area.origin.x = MenuStagger*pane;
      area.origin.y = area.size.height*pane;
      [card setFrame:area];
    }

  card = [list objectAtIndex:pane];
  [card addItem:str value:val enabled:enabled];
  area = [card frame];
  if (area.size.width > maxwidth)
    maxwidth = area.size.width;

  return self;
}


- addTitle:(char *)str to:(int)pane
{
  int row;
  EmacsMenuCard *card;
  NSRect area;
  
  for (row = [list count]; row <= pane; row++)
    {
      card = [[EmacsMenuCard alloc] init];
      [list addObject: card];
      [[self contentView] addSubview:card];

      area.size = [[card cell] cellSize];
      if (area.size.width > maxwidth)  maxwidth=area.size.width;
      area.origin.x = MenuStagger*pane;
      area.origin.y = area.size.height*pane;
      [card setFrame:area];
    }
  
  card = [list objectAtIndex:pane];
  [card addTitle:str];

  area.size = [[card cell] cellSize];
  if (area.size.width > maxwidth) 
    maxwidth=area.size.width;

  return self;
}


- initFromMenu:(Lisp_Object)menu
{
  Lisp_Object title;
  [super init];

  list = [[NSMutableArray alloc] init];
  
  if (!NILP (Fkeymapp(menu)))
    {
      Lisp_Object keymap = get_keymap(menu);
      process_keymap(self, keymap, Qnil, 0);
      title = Fkeymap_prompt(keymap);
    }
  else if (XTYPE(menu)==Lisp_Cons && !NILP(Fkeymapp (Fcar(menu))))
    {
      Lisp_Object tem,keymap;
      for (title=Qnil; XTYPE(menu) == Lisp_Cons; menu = Fcdr(menu))
        {
          keymap = get_keymap(Fcar(menu));
          process_keymap(self, keymap, Qnil, 0);
          tem = Fkeymap_prompt (keymap);
          if (NILP(title) && !NILP(tem)) title = tem;
        }
    }
  else
    {
      title = Fcar(menu);
      process_list(self,Fcdr(menu));
    }
  
  if (XTYPE(title) == Lisp_String)
    {
      NSString *titleStr;

      titleStr = [NSString stringWithUTF8String: XSTRING(title)->data];
      if (titleStr == nil)
          titleStr = @"< ? >";

      [self setTitle: titleStr];
    }
  else
    [self setTitle: @"Select..."];
  
  {
    int i;
    NSRect r,s,t;
    
    if ([list count] > 0)
      {
        t = [(EmacsMenuCard *)[list objectAtIndex: 0] frame];
        for (i = 0; i < [list count]; i++)
          [[list objectAtIndex: i] 
                  setFrameSize: NSMakeSize(maxwidth, t.size.height)];
      }
    else
      {
        t = NSMakeRect(0,0,0,0);
        fprintf(stderr, "WARNING: menu panel w/no contents.\n");
      }
    r = [self frame];
    s = [(NSView *)[self contentView] frame];
    r.size.height += t.size.height*i-s.size.height;
    r.size.width  += maxwidth+(i-1)*MenuStagger-s.size.width;
    [self setFrame: r display: YES];
  }
  
  return self;
}


- (void)dealloc
{
  [list release];
  [super dealloc]; 
  return;
}


- (int)cards
{
  return [list count];
}


- (Lisp_Object)runMenuAt:(float)x:(float)y
{
  NSEvent *e;
  NSModalSession session;
  NSRect r;
  int ret;
  
  if ([list count] == 0)
    return Qnil;
  
  r = [(EmacsMenuCard *)[list objectAtIndex:0] frame];
  r.origin = [[self contentView] convertPoint:r.origin toView:nil];
  
  [self setFrameOrigin:NSMakePoint(x-r.size.width/2.0,
                                   y-r.origin.y+r.size.height/2.0)];
  [self orderFront:NSApp];
  
  session = [NSApp beginModalSessionForWindow:self];

  [self setHidesOnDeactivate:YES];
  while ((ret=[NSApp runModalSession:session])==NSRunContinuesResponse)
    (e = [NSApp nextEventMatchingMask:NSAnyEventMask
                            untilDate:[NSDate distantFuture]
                               inMode:NSEventTrackingRunLoopMode
                              dequeue:(!1)]);
  
  [NSApp endModalSession:session];

  return (Lisp_Object)ret;
}
@end



/* For popup dialogs. */
@implementation EmacsDialogPanel

#define SPACER		8.0
#define ICONSIZE	50.0
#define TEXTHEIGHT	20.0

- initWithContentRect:(NSRect)contentRect styleMask:(unsigned int)aStyle
              backing:(NSBackingStoreType)backingType defer:(BOOL)flag
{
  NSSize spacing = {SPACER,SPACER};
  NSRect area;
  char this_cmd_name[80];
  id cell,tem;
  static NSImageView *imgView;

  if (imgView == nil)
    {
      NSImage *img;
      area.origin.x   = SPACER;
      area.origin.y   = SPACER;
      area.size.width = ICONSIZE;
      area.size.height= ICONSIZE;
      img = [[NSImage imageNamed: @"NSApplicationIcon"] copy];
      [img setScalesWhenResized: YES];
      [img setSize: NSMakeSize(ICONSIZE, ICONSIZE)];
      imgView = [[NSImageView alloc] initWithFrame: area];
      [imgView setImage: img];
      [imgView setEditable: NO];
      [img release];
    }

  aStyle=NSTitledWindowMask;
  flag=YES;
  rows=0;
  cols=1;
  [super initWithContentRect:contentRect styleMask:aStyle
                     backing:backingType defer:flag];
  [[self contentView] setAutoresizesSubviews:YES];

  [[self contentView] addSubview: imgView];
  [self setTitle: @"Emacs"];

  if (XTYPE(Vthis_command)==Lisp_Symbol)
    {
      int i;

      strcpy(this_cmd_name,XSTRING(XSYMBOL(Vthis_command)->xname)->data);
      if (this_cmd_name[0] >= 'a' && this_cmd_name[0] <= 'z')
        this_cmd_name[0] = this_cmd_name[0]-'a'+'A';
      for (i=0; this_cmd_name[i] != '\0'; i++)
        {
          if (this_cmd_name[i] == '-')
            {
              this_cmd_name[i]=' ';
              if (this_cmd_name[i+1] >= 'a' && this_cmd_name[i+1] <= 'z')
                this_cmd_name[i+1] = this_cmd_name[i+1]-'a'+'A';
            }
        }
    }
  else
    strcpy(this_cmd_name,"Emacs");

  area.origin.x   = ICONSIZE+2*SPACER;
  area.origin.y   = ICONSIZE/2-10+SPACER;
  area.size.width = 400;
  area.size.height= TEXTHEIGHT;
  command = [[NSTextField alloc] initWithFrame:area];
  [[self contentView] addSubview:command];
  [command setStringValue:[NSString stringWithUTF8String:this_cmd_name]];
  [command setDrawsBackground:NO];
  [command setBezeled:NO];
  [command setSelectable:NO];
  [command setFont:[NSFont systemFontOfSize:18.0]];

  area.origin.x   = 0;
  area.origin.y   = ICONSIZE+2*SPACER;
  area.size.width = 400;
  area.size.height= 2;
  tem = [[NSBox alloc] initWithFrame:area];
  [[self contentView] addSubview:tem];
  [tem setTitlePosition:NSNoTitle];
  [tem setAutoresizingMask:NSViewWidthSizable];

  area.origin.x   = 2*SPACER;
  area.origin.y  += 2*SPACER;
  area.size.width = 400;
  area.size.height= TEXTHEIGHT;
  title = [[NSTextField alloc] initWithFrame:area];
  [[self contentView] addSubview:title];
  [title setDrawsBackground:NO];
  [title setBezeled:NO];
  [title setSelectable:NO];
  [title setFont:[NSFont systemFontOfSize:14.0]];

  cell = [[NSButtonCell alloc] initTextCell:@""];
  // Make it look like nothing's here
  [cell setBordered:NO];
  [cell setEnabled:NO];
  [cell setCellAttribute:NSCellIsInsetButton to:8];

  matrix = [[NSMatrix alloc] initWithFrame:contentRect 
                                      mode:NSHighlightModeMatrix 
                                 prototype:cell 
                              numberOfRows:0 
                           numberOfColumns:1];
  [[self contentView] addSubview:matrix];
  [matrix setFrameOrigin:NSMakePoint(SPACER, area.origin.y+TEXTHEIGHT+2*SPACER)];
  [matrix setIntercellSpacing:spacing];

  [self setOneShot:YES];
  [self setReleasedWhenClosed:YES];
  [self setHidesOnDeactivate:YES];
  return self;
}


- (BOOL)windowShouldClose:(id)sender
{
  [NSApp stopModalWithCode:Qnil];
  return NO;
}


void process_dialog(id window,Lisp_Object list)
{
  Lisp_Object item;
  int row;

  row=0;
  for (; XTYPE(list)==Lisp_Cons; list=XCDR(list))
    {
      item=XCAR(list);
      if (XTYPE(item)==Lisp_String)
        {
          [window addString:XSTRING(item)->data row:row];
          row++;
        }
      else if (XTYPE(item)==Lisp_Cons)
        {
          [window addButton:XSTRING(XCAR(item))->data
                      value:XCDR(item) row:row];
          row++;
        }
      else if (NILP(item))
        {
          [window addSplit];
          row=0;
        }
    }
}


- addButton:(char *)str value:(Lisp_Object)val row:(int)row
{
  id cell;
       
  if (row >= rows)
    {
      [matrix addRow];
      rows++;
    }
  cell = [matrix cellAtRow:row column:cols-1];
  [cell setTarget:self];
  [cell setAction:@selector(clicked:)];
  [cell setTitle:[NSString stringWithUTF8String:str]];
  [cell setTag:(int)val];
  [cell setBordered:YES];
  [cell setEnabled:YES];

  return self;
}


- addString:(char *)str row:(int)row
{
  id cell;
       
  if (row >= rows)
    {
      [matrix addRow];
      rows++;
    }
  cell = [matrix cellAtRow:row column:cols-1];
  // No setAction here.  Disabled buttons can't trigger actions
  [cell setTitle:[NSString stringWithUTF8String:str]];
  [cell setBordered:YES];
  [cell setEnabled:NO];

  return self;
}


- addSplit
{
  [matrix addColumn];
  cols++;
  return self;
}


- clicked:sender
{
  NSArray *sellist=nil;
  Lisp_Object seltag;

  sellist=[sender selectedCells];
  if ([sellist count]<1) 
    return self;

  seltag = (Lisp_Object)[[sellist objectAtIndex:0] tag];
  if (! EQ(seltag,Qundefined))
    [NSApp stopModalWithCode:seltag];
  return self;
}


- initFromContents:(Lisp_Object)contents isQuestion:(BOOL)isQ
{
  Lisp_Object head;
  [super init];

  if (XTYPE(contents)==Lisp_Cons)
    {
      head = Fcar(contents);
      process_dialog(self,Fcdr(contents));
    }
  else
    head = contents;

  if (XTYPE(head)==Lisp_String)
      [title setStringValue:
                 [NSString stringWithUTF8String:XSTRING(head)->data]];
  else if (isQ == YES)
      [title setStringValue: @"Question"];
  else
      [title setStringValue: @"Information"];

  {
    int i;
    NSRect r,s,t;

    if (cols == 1 && rows > 1)	// Never told where to split
      {
        [matrix addColumn];
        for (i=0; i<rows/2; i++)
          {
            [matrix putCell:[matrix cellAtRow:(rows+1)/2 column:0]
                      atRow:i column:1];
            [matrix removeRow:(rows+1)/2];
          }
      }
    [matrix sizeToFit];
    [title sizeToFit];
    [command sizeToFit];

    t = [matrix frame];
    r = [title frame];
    if (r.size.width+r.origin.x > t.size.width+t.origin.x)
      {
        t.origin.x   = r.origin.x;
        t.size.width = r.size.width;
      }
    r = [command frame];
    if (r.size.width+r.origin.x > t.size.width+t.origin.x)
      {
        t.origin.x   = r.origin.x;
        t.size.width = r.size.width;
      }

    r = [self frame];
    s = [(NSView *)[self contentView] frame];
    r.size.height += t.origin.y+t.size.height+SPACER-s.size.height;
    r.size.width  += t.origin.x+t.size.width +SPACER-s.size.width;
    [self setFrame:r display:NO];
  }

  return self;
}


- (void)dealloc
{
  [matrix release];
  { [super dealloc]; return; };
}


- (Lisp_Object)runDialogAt:(float)x:(float)y
{
  NSEvent *e;
  NSModalSession session;
  int ret;

  [self center];
  [self orderFront:NSApp];

  session = [NSApp beginModalSessionForWindow:self];
  while ((ret=[NSApp runModalSession:session])==NSRunContinuesResponse)
    (e = [NSApp nextEventMatchingMask:NSAnyEventMask
                            untilDate:[NSDate distantFuture]
                               inMode:NSEventTrackingRunLoopMode
                              dequeue:NO]);
  [NSApp endModalSession:session];

  return (Lisp_Object)ret;
}

@end


/* Needed because NeXTstep does not provide enough control over tooltip
   display. */
@implementation EmacsTooltip

- init
{
  NSColor *col = [NSColor colorWithCalibratedRed: 1.0 green: 1.0
                                            blue: 0.792 alpha: 0.9];
  NSFont *font = [NSFont toolTipsFontOfSize: 0];
  NSFont *sfont = [font screenFont];
  int height = [sfont ascender] - [sfont descender];//[font boundingRectForFont].size.height;
  NSRect r = NSMakeRect(0, 0,100, height+6);

  textField = [[NSTextField alloc] initWithFrame: r];
  [textField setFont: font];
  [textField setBackgroundColor: col];

  [textField setEditable: NO];
  [textField setSelectable: NO];
  [textField setBordered: YES];
  [textField setBezeled: NO];
  [textField setDrawsBackground: NO];

  win = [[NSWindow alloc]
            initWithContentRect: [textField frame]
                      styleMask: 0
                        backing: NSBackingStoreBuffered
                          defer: YES];
  [win setReleasedWhenClosed: NO];
  [win setDelegate:self];
  [[win contentView] addSubview: textField];
  [win setBackgroundColor: col];
  [win setOpaque: NO];

  return self;
}

- (void) dealloc
{
  [win close];
  [win release];
  [textField release];
  [super dealloc];
}

- (void) setText: (char *)text
{
  NSString *str = [NSString stringWithUTF8String: text];
  NSRect r = [textField frame];
  r.size.width = [[[textField font] screenFont] widthOfString: str] + 8;
  [textField setFrame: r];
  [textField setStringValue: str];
}

- (void) showAtX: (int)x Y: (int)y for: (int)seconds
{
  NSRect wr = [win frame];

  wr.origin = NSMakePoint(x, y);
  wr.size = [textField frame].size;

  [win setFrame: wr display: YES];
  [win orderFront: self];
  [win display];
  timer = [NSTimer scheduledTimerWithTimeInterval: (float)seconds target: self
                                         selector: @selector(hide) userInfo:nil
                                          repeats: NO];
  [timer retain];
}

- (void) hide
{
  [win close];
  if (timer != nil)
    {
      if ([timer isValid])
        [timer invalidate];
      [timer release];
      timer = nil;
    }
}

- (BOOL) isActive
{
  return timer != nil;
}

- (NSRect) frame
{
  return [textField frame];
}

@end



// ============================================================================
//
//    Toolbar: externally-called functions
//
// ============================================================================

void
free_frame_tool_bar (FRAME_PTR f)
// ----------------------------------------------------------------------------
//    Under NS we just hide the toolbar until it might be needed again.
// ----------------------------------------------------------------------------
{
  [[FRAME_NS_VIEW(f) toolbar] setVisible: NO];
}

void
update_frame_tool_bar (FRAME_PTR f)
// ----------------------------------------------------------------------------
//    Update toolbar contents
// ----------------------------------------------------------------------------
{
  int i;
  EmacsToolbar *toolbar = [FRAME_NS_VIEW(f) toolbar];
//fprintf(stderr, "update toolbar\n");

 if (NILP(f->tool_bar_lines) || !INTEGERP(f->tool_bar_lines))
   return;

  [toolbar clearActive];

  // update EmacsToolbar as in GtkUtils, build items list
  for (i = 0; i < f->n_tool_bar_items; ++i)
    {
#define TOOLPROP(IDX) AREF (f->tool_bar_items, i * TOOL_BAR_ITEM_NSLOTS + (IDX))

      BOOL enabled_p = !NILP (TOOLPROP (TOOL_BAR_ITEM_ENABLED_P));
      BOOL selected_p = !NILP (TOOLPROP (TOOL_BAR_ITEM_SELECTED_P));
      int idx;
      int img_id;
      struct image *img;
      Lisp_Object image;
      Lisp_Object helpObj;
      char *helpText;

      /* If image is a vector, choose the image according to the
	 button state.  */
      image = TOOLPROP (TOOL_BAR_ITEM_IMAGES);
      if (VECTORP (image))
	{
	  if (enabled_p)
              idx = (selected_p
		   ? TOOL_BAR_IMAGE_ENABLED_SELECTED
		   : TOOL_BAR_IMAGE_ENABLED_DESELECTED);
	  else
              idx = (selected_p
		   ? TOOL_BAR_IMAGE_DISABLED_SELECTED
		   : TOOL_BAR_IMAGE_DISABLED_DESELECTED);

	  xassert (ASIZE (image) >= idx);
	  image = AREF (image, idx);
	}
      else
        {
          idx = -1;
        }
      /* Ignore invalid image specifications.  */
      if (!valid_image_p (image))
        {
          NSLog(@"Invalid image for toolbar item");
          continue;
        }

      img_id = lookup_image (f, image);
      img = IMAGE_FROM_ID (f, img_id);
      prepare_image_for_display (f, img);

      if (img->load_failed_p || img->pixmap == nil)
        {
          NSLog(@"Could not prepare toolbar image for display.");
          continue;
        }

      helpObj = TOOLPROP(TOOL_BAR_ITEM_HELP);
      if (NILP(helpObj))
        helpObj = TOOLPROP(TOOL_BAR_ITEM_CAPTION);
      helpText = NILP(helpObj) ? "" : (char *)SDATA(helpObj);

      [toolbar addDisplayItemWithImage: img->pixmap idx: i helpText: helpText
                               enabled: enabled_p];
#undef TOOLPROP
    }

  if (![toolbar isVisible])
      [toolbar setVisible: YES];

  if ([toolbar changed])
    {
//fprintf(stderr, "Toolbar changed\n");
      // inform app that toolbar has changed
      NSDictionary *dict = [toolbar configurationDictionary];
      NSMutableDictionary *newDict = [dict mutableCopy];
      NSEnumerator *keys = [[dict allKeys] objectEnumerator];
      NSObject *key;
      while ((key = [keys nextObject]) != nil)
        {
          NSObject *val = [dict objectForKey: key];
          if ([val isKindOfClass: [NSArray class]])
            {
              [newDict setObject:
                         [toolbar toolbarDefaultItemIdentifiers: toolbar]
                          forKey: key];
              break;
            }
        }
      [toolbar setConfigurationFromDictionary: newDict];
    }

}



// ============================================================================
//
//    Toolbar: class implementations
//
// ============================================================================

@implementation EmacsToolbar

- initForView: (EmacsView *)view withIdentifier: (NSString *)identifier
{
  self = [super initWithIdentifier: identifier];
  emacsView = view;
  [self setDisplayMode: NSToolbarDisplayModeIconOnly];
  [self setSizeMode: NSToolbarSizeModeSmall];
  [self setDelegate: self];
  identifierToItem = [[NSMutableDictionary alloc] initWithCapacity: 10];
  activeIdentifiers = [[NSMutableArray alloc] initWithCapacity: 8];
  return self;
}

- (void)dealloc
{
  [prevIdentifiers release];
  [activeIdentifiers release];
  [identifierToItem release];
  [super dealloc];
}

- (void) clearActive
{
  [prevIdentifiers release];
  prevIdentifiers = [activeIdentifiers copy];
  [activeIdentifiers removeAllObjects];
}

- (BOOL) changed
{
  return [activeIdentifiers isEqualToArray: prevIdentifiers] ? NO : YES;
}

- (void) addDisplayItemWithImage: (EmacsImage *)img idx: (int)idx
                        helpText: (char *)help enabled:(BOOL)enabled
{
  // 1) come up w/identifier
  NSString *identifier =
    [NSString stringWithFormat: @"%u", [img hash]];
//    [NSString stringWithFormat: @"%u %d", [img hash], enabled];

  // 2) create / reuse item
  NSToolbarItem *item = [identifierToItem objectForKey: identifier];
  if (item == nil)
    {
      item = [[[NSToolbarItem alloc] initWithItemIdentifier: identifier] autorelease];
      [item setImage: img];
      [item setToolTip: [NSString stringWithCString: help]];
      [item setTarget:emacsView];
      [item setAction:@selector(toolbarClicked:)];
//NSLog(@"creating toolbar item '%@'", identifier);
    }

  [item setTag: idx];
  [item setEnabled: enabled];

  // 3) update state
  [identifierToItem setObject: item forKey: identifier];
  [activeIdentifiers addObject: identifier];
}

// This overrides super's implementation, which automatically sets
// all items to enabled state (for some reason).
- (void)validateVisibleItems { }


// delegate methods

- (NSToolbarItem *)toolbar: (NSToolbar *)toolbar
      itemForItemIdentifier:(NSString *)itemIdentifier
  willBeInsertedIntoToolbar:(BOOL)flag
{
  // look up NSToolbarItem by identifier and return...
  return [identifierToItem objectForKey: itemIdentifier];
}

- (NSArray *)toolbarDefaultItemIdentifiers: (NSToolbar *)toolbar
{
  // return entire set..
  return activeIdentifiers;
}

// for configuration palette (not yet supported)
- (NSArray *)toolbarAllowedItemIdentifiers: (NSToolbar *)toolbar
{
  // return entire set...
  return [identifierToItem allKeys];
}

// optional and unneeded
//- toolbarWillAddItem: (NSNotification *)notification { }
//- toolbarDidRemoveItem: (NSNotification *)notification { }
//- (NSArray *)toolbarSelectableItemIdentifiers: (NSToolbar *)toolbar

@end
